Type: Package
Package: autosync
Title: 'Automerge' Sync Server and Client
Version: 0.1.0
Authors@R: c(
    person("Charlie", "Gao", , "charlie.gao@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-0750-061X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: A WebSocket-based implementation of the 'automerge-repo'
    synchronization protocol used by 'sync.automerge.org'. Acts as a sync
    server, enabling 'R' to serve as a synchronization hub for 'Automerge'
    clients in 'JavaScript', 'Rust', and other languages, and as a client
    for fetching, editing, and synchronizing documents hosted on remote
    servers.
License: MIT + file LICENSE
URL: https://posit-dev.github.io/autosync/,
        https://github.com/posit-dev/autosync
BugReports: https://github.com/posit-dev/autosync/issues
Depends: R (>= 4.4)
Imports: automerge (>= 0.4.0), httr2 (>= 1.2.3), jose, later, nanonext
        (>= 1.8.1), promises, secretbase (>= 1.3.0)
Suggests: openssl, testthat (>= 3.0.0)
Config/Needs/website: tidyverse/tidytemplate
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0.9000
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-06-30 20:04:11 UTC; cg334
Author: Charlie Gao [aut, cre] (ORCID: <https://orcid.org/0000-0002-0750-061X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Charlie Gao <charlie.gao@posit.co>
Repository: CRAN
Date/Publication: 2026-07-06 14:00:10 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,autosync_client)
S3method(print,autosync_doc)
S3method(print,autosync_server)
export(auth_config)
export(create_document)
export(generate_document_id)
export(get_document)
export(list_documents)
export(sync_client)
export(sync_fetch)
export(sync_server)
export(sync_token)
importFrom(automerge,
  AM_ROOT,
  am_create,
  am_get,
  am_keys,
  am_length,
  am_load,
  am_save,
  am_sync_decode,
  am_sync_encode,
  am_sync_state,
  am_sync_state_decode,
  am_sync_state_encode
)
importFrom(httr2,
  oauth_client,
  oauth_flow_auth_code,
  oauth_redirect_uri,
  oauth_server_metadata
)
importFrom(jose,
  jwt_decode_sig,
  read_jwk
)
importFrom(later,
  later,
  run_now
)
importFrom(nanonext,
  handler_ws,
  http_server,
  is_error_value,
  ncurl,
  random,
  recv,
  recv_aio,
  send,
  stop_aio,
  stream,
  tls_config,
  unresolved,
  write_cert
)
importFrom(promises,then)
importFrom(secretbase,
  base58dec,
  base58enc,
  base64dec,
  base64enc,
  cbordec,
  cborenc,
  jsondec,
  jsonenc
)
importFrom(utils,str)

# autosync

<!-- badges: start -->

[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental) [![R-CMD-check](https://github.com/posit-dev/autosync/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/posit-dev/autosync/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/posit-dev/autosync/graph/badge.svg)](https://app.codecov.io/gh/posit-dev/autosync)
<!-- badges: end -->

R sync server and client for [Automerge](https://automerge.org/) CRDT documents. Implements the `automerge-repo` WebSocket protocol, enabling R to serve as a synchronization hub for Automerge clients in R, JavaScript, Rust, and other languages, and to act as a client that fetches, edits, and synchronizes documents hosted on remote servers.

[![Ask DeepWiki](https://deepwiki.com/badge.svg)](https://deepwiki.com/posit-dev/autosync)

### Features

- Full `automerge-repo` protocol compatibility (interoperates with sync.automerge.org)
- Secure connections via TLS (wss://)
- Persistent document storage
- Ephemeral message broadcasting
- Client functions for fetching documents from remote servers

## Installation

Install the released version from CRAN:

``` r
install.packages("autosync")
```

Or the development version from GitHub:

``` r
pak::pak("posit-dev/autosync")
```

## Server

Create a WebSocket sync server:

``` r
library(autosync)

server <- sync_server()
server$start()
server

# Server runs non-blocking in the background

server$close()
```

With TLS for secure connections:

``` r
cert <- nanonext::write_cert()
tls <- nanonext::tls_config(server = cert$server)

server <- sync_server(tls = tls)
server$start()
server$url
```

### Authentication

Enable OIDC authentication to restrict access. The defaults point to
Google, but any OIDC-compliant provider (Microsoft Entra, Okta, Auth0,
etc.) can be used by setting the `issuer` argument.

Register your OAuth client as a **"Desktop app"** and set the
`OIDC_CLIENT_ID` and `OIDC_CLIENT_SECRET` environment variables (e.g. in
`.Renviron`):

``` r
# .Renviron
OIDC_CLIENT_ID=YOUR_CLIENT_ID
OIDC_CLIENT_SECRET=YOUR_CLIENT_SECRET
```

Google issues a client secret for Desktop app clients and requires it in
the token exchange, but — unlike a "Web application" secret — it is not
treated as confidential: it is designed to be distributed with installed
apps, so it is safe to keep in `.Renviron`. Google states that for
installed apps "the client secret is obviously not treated as a secret"
([Google OAuth 2.0 docs][oauth-google]), consistent with the OAuth 2.0
for Native Apps standard ([RFC 8252 §8.5][rfc8252-85]). The interactive
flow also uses PKCE, so the secret alone does not grant access.

[oauth-google]: https://developers.google.com/identity/protocols/oauth2#installed
[rfc8252-85]: https://datatracker.ietf.org/doc/html/rfc8252#section-8.5

Other providers (Microsoft Entra, Okta, Auth0, etc.) can register the
client as a **native / public client**, which authenticates via PKCE
alone — for those, set only `OIDC_CLIENT_ID` and leave
`OIDC_CLIENT_SECRET` unset.

``` r
cert <- nanonext::write_cert()
tls <- nanonext::tls_config(server = cert$server)

server <- sync_server(
  tls = tls,  # TLS required when auth is enabled
  auth = auth_config(allowed_domains = "posit.co")
)
server$start()
```

Clients obtain a token interactively, then pass it when fetching:

``` r
token <- sync_token()

tlsclient <- nanonext::tls_config(client = cert$client)
doc <- sync_fetch(
  server$url,
  "35ei6ouA7nLhtjmf3d9xk1KKvtKv",
  token = token,
  tls = tlsclient
)
```

### Sharing policy

The `share` parameter controls which documents are announced to clients and which requests are allowed. It accepts `TRUE`, `FALSE`, `NA`, or a function returning one of these per client and document:

| Value | Announce | Allow requests |
|-------|----------|----------------|
| `NA` (default) | No | Yes |
| `TRUE` | Yes | Yes |
| `FALSE` | No | No |

Announce all documents to every client that connects:

``` r
server <- sync_server(share = TRUE)
```

Combine with authentication and use a `share` function for fine-grained access control. The function receives the `client_id` and `doc_id`, and can look up the authenticated email on the connection to decide:

``` r
# Allow list of emails that can access documents
allowed <- c("alice@example.com", "bob@example.com")

server <- sync_server(
  tls = tls,
  auth = auth_config(),
  share = function(client_id, doc_id) {
    state <- attr(server, "sync")
    conn <- state$connections[[client_id]]
    if (conn$authenticated_email %in% allowed) NA else FALSE
  }
)
```

### Document management

``` r
# Create a new document
doc_id <- create_document(server)

# List all documents
list_documents(server)

# Get a document
doc <- get_document(server, doc_id)
```

### Connecting from JavaScript

``` javascript
import { Repo } from "@automerge/automerge-repo"
import { BrowserWebSocketClientAdapter } from "@automerge/automerge-repo-network-websocket"

const repo = new Repo({
  network: [new BrowserWebSocketClientAdapter("ws://localhost:8080")]
})
```

## Client

Fetch a document from any automerge-repo sync server in a single,
one-off retrieval over a throwaway connection:

``` r
# Fetch from public sync server
doc <- sync_fetch("wss://sync.automerge.org", "your-document-id")

# Inspect document structure
str(doc)

# Or verbose fetch for debugging sync issues
doc <- sync_fetch("wss://sync.automerge.org", "your-document-id", verbose = TRUE)
```

### Live connections

For a persistent connection that keeps documents in sync — receiving
real-time updates from other peers and flushing local changes — use
`sync_client()`. Several documents can share a single connection, each
opened with `$open_doc()`:

``` r
conn <- sync_client("wss://sync.automerge.org")

# Open a live document over the connection
doc <- conn$open_doc("your-document-id")
automerge::am_keys(doc$doc)

# Make local changes and push them to the server
automerge::am_put(doc$doc, automerge::AM_ROOT, "key", "value")
doc$push()

# Open another document over the same connection
other <- conn$open_doc("another-document-id")

# Disconnect (closes every document on the connection)
conn$close()
```

## Utilities

Generate document IDs compatible with automerge-repo:

``` r
doc_id <- generate_document_id()
```

## Links

- [automerge](https://posit-dev.github.io/automerge-r/) - R bindings for Automerge CRDT

# Automerge sync server implementation using nanonext

#' Create an Automerge sync server
#'
#' Creates a WebSocket server that implements the automerge-repo sync protocol,
#' compatible with JavaScript, Rust, and other Automerge clients.
#'
#' @param port Port to listen on. Default 0 (binds to a random available port).
#'   The actual URL is retrieved via `server$url`.
#' @param host Host address to bind to. Default "127.0.0.1" (localhost).
#' @param data_dir Directory for document storage. Defaults to a
#'   session-temporary directory, so documents do not persist across R
#'   sessions. Supply an explicit path to persist documents across sessions.
#' @param auto_create_docs Logical, whether to auto-create documents when
#'   clients request unknown document IDs. Default TRUE.
#' @param storage_id Optional storage ID for this server. If NULL (default),
#'   generates a new persistent identity. Set to NA for an ephemeral server
#'   (no persistence identity).
#' @param tls (optional) for secure wss:// connections, a TLS configuration
#'   object created by [nanonext::tls_config()].
#' @param auth Optional authentication configuration created by [auth_config()].
#'   When provided, clients must include a valid JWT (ID token) as a
#'   Bearer token in the Authorization header of the WebSocket upgrade request.
#'   Connections without valid credentials are rejected immediately.
#'   Note: TLS is required when authentication is enabled to protect tokens.
#' @param share Controls document sharing policy for connected clients. This
#'   unified parameter governs both proactive document announcement (pushing
#'   documents to clients) and access control (allowing clients to request
#'   documents). Accepts one of:
#'   \itemize{
#'     \item `NA` (default) — never announce but allow all requests.
#'     \item `TRUE` — announce all documents to all clients and allow all
#'       requests.
#'     \item `FALSE` — never announce and deny all requests (sends
#'       `doc-unavailable`).
#'     \item A function with signature `function(client_id, doc_id)`
#'       returning `TRUE` (announce and allow), `NA` (allow on request only),
#'       or `FALSE` (deny access). Called per client and per document.
#'   }
#'
#' @return An autosync_server object inheriting from 'nanoServer', with
#'   `$start()` and `$close()` methods.
#'
#' @details
#' The returned server inherits from nanonext's nanoServer class and provides
#' `$start()` and `$close()` methods for non-blocking operation.
#'
#' @examplesIf interactive()
#' # Create and start a server
#' server <- sync_server()
#' server$start()
#'
#' # Server is now running in the background
#' # ...do other work...
#'
#' # Stop when done
#' server$close()
#'
#' # With TLS for secure connections
#' cert <- nanonext::write_cert()
#' tls <- nanonext::tls_config(server = cert$server)
#' server <- sync_server(tls = tls)
#' server$start()
#' server$url
#' server$close()
#'
#' # Server with OIDC authentication (requires TLS)
#' cert <- nanonext::write_cert()
#' tls <- nanonext::tls_config(server = cert$server)
#' server <- sync_server(
#'   tls = tls,
#'   auth = auth_config(
#'     client_id = "123456789.apps.googleusercontent.com",
#'     allowed_domains = "mycompany.com"
#'   )
#' )
#'
#' @export
sync_server <- function(
  port = 0L,
  host = "127.0.0.1",
  data_dir = tempfile("autosync"),
  auto_create_docs = TRUE,
  storage_id = NULL,
  tls = NULL,
  auth = NULL,
  share = NA
) {
  port <- as.integer(port)

  # Enforce TLS when authentication is enabled
  if (!is.null(auth) && is.null(tls)) {
    stop(
      "Authentication requires TLS. Provide a 'tls' configuration.\n",
      "Transmitting tokens over unencrypted connections is a security risk."
    )
  }

  scheme <- if (is.null(tls)) "ws" else "wss"
  url <- sprintf("%s://%s:%d", scheme, host, port)

  documents <- new.env(hash = TRUE, parent = emptyenv())
  sync_states <- new.env(hash = TRUE, parent = emptyenv())
  connections <- new.env(hash = TRUE, parent = emptyenv())
  doc_peers <- new.env(hash = TRUE, parent = emptyenv())

  if (!dir.exists(data_dir)) {
    dir.create(data_dir, recursive = TRUE)
  }

  # Create state environment for handlers to access via closure

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$host <- host
  state$data_dir <- data_dir
  state$auto_create_docs <- auto_create_docs
  state$peer_id <- generate_peer_id()
  state$storage_id <- if (is.null(storage_id)) {
    generate_peer_id()
  } else if (is.na(storage_id)) {
    NULL
  } else {
    storage_id
  }
  state$documents <- documents
  state$sync_states <- sync_states
  state$connections <- connections
  state$doc_peers <- doc_peers
  state$ephemeral_counts <- new.env(hash = TRUE, parent = emptyenv())
  state$auth <- auth
  state$share <- share

  load_all_documents(state)

  on_open <- function(ws, req) {
    ws_id <- as.character(ws$id)

    # Authenticate from Authorization header if auth is enabled
    if (!is.null(state$auth)) {
      auth_result <- authenticate_header(state$auth, req$headers)
      if (!auth_result$valid) {
        ws$send(cborenc(list(
          type = "error",
          senderId = state$peer_id,
          message = auth_result$error
        )))
        ws$close()
        return(invisible())
      }
    }

    state$connections[[ws_id]] <- list(
      ws = ws,
      client_id = NULL,
      metadata = NULL,
      connected_at = Sys.time(),
      authenticated_email = if (!is.null(state$auth)) auth_result$email
    )
  }

  on_message <- function(ws, data) {
    ws_id <- as.character(ws$id)
    conn <- state$connections[[ws_id]]
    client_id <- if (!is.null(conn$client_id)) conn$client_id else ws_id
    handle_message(state, client_id, ws_id, data)
  }

  on_close <- function(ws) {
    ws_id <- as.character(ws$id)
    conn <- state$connections[[ws_id]]
    if (!is.null(conn)) {
      client_id <- conn$client_id
      handle_disconnect(state, client_id)
      rm(list = ws_id, envir = state$connections)
      if (!is.null(client_id) && exists(client_id, envir = state$connections)) {
        rm(list = client_id, envir = state$connections)
      }
    }
  }

  ws_handler <- handler_ws(
    path = "/",
    on_message = on_message,
    on_open = on_open,
    on_close = on_close,
    textframes = FALSE
  )

  server <- http_server(
    url = url,
    handlers = list(ws_handler),
    tls = tls
  )

  attr(server, "sync") <- state
  class(server) <- c("autosync_server", class(server))

  server
}

#' Get a document from the server
#'
#' Retrieves an Automerge document by its ID.
#'
#' @param server An autosync_server object.
#' @param doc_id Document ID string.
#'
#' @return Automerge document object, or NULL if not found.
#'
#' @examplesIf interactive()
#' server <- sync_server()
#' doc_id <- create_document(server)
#' get_document(server, doc_id)
#' server$close()
#'
#' @export
get_document <- function(server, doc_id) {
  attr(server, "sync")$documents[[doc_id]]
}

#' List all document IDs
#'
#' Returns the IDs of all documents currently loaded in the server.
#'
#' @param server An autosync_server object.
#'
#' @return Character vector of document IDs.
#'
#' @examplesIf interactive()
#' server <- sync_server()
#' create_document(server)
#' list_documents(server)
#' server$close()
#'
#' @export
list_documents <- function(server) {
  ls(attr(server, "sync")$documents)
}

#' Create a new document on the server
#'
#' Creates a new empty Automerge document and registers it with the server.
#'
#' @param server An autosync_server object.
#' @param doc_id Optional document ID. If NULL, generates a new ID.
#'
#' @return Document ID string.
#'
#' @examplesIf interactive()
#' server <- sync_server()
#' doc_id <- create_document(server)
#' server$close()
#'
#' @export
create_document <- function(server, doc_id = NULL) {
  state <- attr(server, "sync")

  if (is.null(doc_id)) {
    doc_id <- generate_document_id()
  }

  doc <- am_create()
  state$documents[[doc_id]] <- doc
  save_document(state, doc_id, doc)
  announce_new_document(state, doc_id, doc)

  doc_id
}

#' Print method for autosync_server
#'
#' @param x An autosync_server object.
#' @param ... Ignored.
#'
#' @return Invisibly returns x.
#'
#' @keywords internal
#' @export
print.autosync_server <- function(x, ...) {
  state <- attr(x, "sync")
  cat("Automerge Sync Server\n")
  cat("  URL:", x$url, "\n")
  cat("  Data dir:", state$data_dir, "\n")
  cat("  Documents:", length(state$documents), "\n")
  cat("  Connections:", length(state$connections), "\n")
  NextMethod()
}

# Integration tests that don't require async callbacks

test_that("document persistence survives server restart", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  # Create server, add document
  server1 <- sync_server(data_dir = data_dir)

  doc_id <- create_document(server1)
  doc <- get_document(server1, doc_id)
  automerge::am_put(doc, automerge::AM_ROOT, "persistent", "data")
  autosync:::save_document(attr(server1, "sync"), doc_id, doc)

  server1$close()

  # Create new server with same data_dir
  server2 <- sync_server(data_dir = data_dir)
  on.exit(server2$close(), add = TRUE)

  # Document should be loaded
  expect_true(doc_id %in% list_documents(server2))

  doc2 <- get_document(server2, doc_id)
  expect_equal(automerge::am_get(doc2, automerge::AM_ROOT, "persistent"), "data")
})

test_that("server creates data directory if it doesn't exist", {
  data_dir <- file.path(tempfile(), "nested", "path")
  on.exit(unlink(dirname(dirname(data_dir)), recursive = TRUE))

  expect_false(dir.exists(data_dir))

  server <- sync_server(data_dir = data_dir)
  on.exit(server$close(), add = TRUE)

  expect_true(dir.exists(data_dir))
})

test_that("multiple documents can be managed simultaneously", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  on.exit(server$close(), add = TRUE)

  # Create multiple documents
  doc_ids <- sapply(1:5, function(i) {
    doc_id <- create_document(server)
    doc <- get_document(server, doc_id)
    automerge::am_put(doc, automerge::AM_ROOT, "index", i)
    doc_id
  })

  # Verify each document
  for (i in 1:5) {
    doc <- get_document(server, doc_ids[i])
    expect_equal(automerge::am_get(doc, automerge::AM_ROOT, "index"), i)
  }
})

test_that("get_document returns NULL for non-existent document", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  on.exit(server$close())

  result <- get_document(server, "nonexistent123")
  expect_null(result)
})

test_that("create_document with explicit ID", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  on.exit(server$close())

  explicit_id <- generate_document_id()
  returned_id <- create_document(server, doc_id = explicit_id)

  expect_equal(returned_id, explicit_id)
  expect_true(explicit_id %in% list_documents(server))
})

test_that("server URL format is correct", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(host = "127.0.0.1", data_dir = data_dir)
  on.exit(server$close())

  expect_true(grepl("^ws://127\\.0\\.0\\.1:\\d+$", server$url))
})

test_that("save_sync_state and load_sync_states round-trip correctly", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$data_dir <- data_dir
  state$sync_states <- new.env(hash = TRUE, parent = emptyenv())

  storage_id <- "testStoragePeer"
  doc_id <- generate_document_id()
  sync_state <- automerge::am_sync_state()

  # Save
  autosync:::save_sync_state(state, storage_id, doc_id, sync_state)
  expect_true(file.exists(
    file.path(data_dir, ".sync_states", storage_id, paste0(doc_id, ".sync"))
  ))

  # Load into a new client_id
  autosync:::load_sync_states(state, storage_id, "newClientId")

  expect_true(exists("newClientId", envir = state$sync_states))
  expect_true(exists(doc_id, envir = state$sync_states[["newClientId"]]))
  expect_true(inherits(state$sync_states[["newClientId"]][[doc_id]], "am_syncstate"))
})

test_that("load_sync_states with no persisted data is a no-op", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$data_dir <- data_dir
  state$sync_states <- new.env(hash = TRUE, parent = emptyenv())

  autosync:::load_sync_states(state, "nonexistentPeer", "clientX")

  expect_false(exists("clientX", envir = state$sync_states))
})

test_that("save_sync_state warns when the file cannot be written", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$data_dir <- data_dir

  storage_id <- "writeFailPeer"
  doc_id <- "blockedDoc"
  # Pre-create a *directory* exactly where the .sync file would be written, so
  # writeBin() fails on it.
  blocking_dir <- file.path(
    data_dir, ".sync_states", storage_id, paste0(doc_id, ".sync")
  )
  dir.create(blocking_dir, recursive = TRUE)

  # writeBin() to a directory emits incidental platform warnings before the
  # connection error that save_sync_state() catches; muffle those so only the
  # "Failed to save sync state" warning we care about is asserted.
  expect_warning(
    withCallingHandlers(
      autosync:::save_sync_state(state, storage_id, doc_id, automerge::am_sync_state()),
      warning = function(w) {
        if (!grepl("Failed to save sync state", conditionMessage(w))) {
          invokeRestart("muffleWarning")
        }
      }
    ),
    "Failed to save sync state"
  )
})

test_that("load_sync_states ignores a peer directory with no sync files", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$data_dir <- data_dir
  state$sync_states <- new.env(hash = TRUE, parent = emptyenv())

  storage_id <- "emptyPeer"
  dir.create(file.path(data_dir, ".sync_states", storage_id), recursive = TRUE)

  autosync:::load_sync_states(state, storage_id, "clientE")

  expect_false(exists("clientE", envir = state$sync_states))
})

test_that("load_sync_states warns on a corrupted sync file", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$data_dir <- data_dir
  state$sync_states <- new.env(hash = TRUE, parent = emptyenv())

  storage_id <- "corruptPeer"
  dir <- file.path(data_dir, ".sync_states", storage_id)
  dir.create(dir, recursive = TRUE)
  doc_id <- generate_document_id()
  writeBin(as.raw(c(0xFF, 0xFF, 0xFF)), file.path(dir, paste0(doc_id, ".sync")))

  expect_warning(
    autosync:::load_sync_states(state, storage_id, "clientC"),
    "Failed to load sync state"
  )
})

test_that("sync_server creates valid server object", {
  server <- sync_server(data_dir = tempdir())
  on.exit(server$close())

  state <- attr(server, "sync")
  expect_s3_class(server, "autosync_server")
  expect_type(state$peer_id, "character")
  expect_type(state$storage_id, "character")
  expect_true(is.environment(state$documents))
  expect_true(is.environment(state$sync_states))
  expect_true(is.environment(state$connections))
})

test_that("sync_server with TLS creates wss URL", {
  server <- sync_server(
    port = 3030L,
    tls = nanonext::tls_config(server = nanonext::write_cert()$server),
    data_dir = tempdir()
  )
  on.exit(server$close())

  expect_true(grepl("^wss://", server$url))
})

test_that("sync_server without TLS creates ws URL", {
  server <- sync_server(data_dir = tempdir())
  on.exit(server$close())

  expect_true(grepl("^ws://", server$url))
})

test_that("sync_server with ephemeral storage_id", {
  server <- sync_server(storage_id = NA, data_dir = tempdir())
  on.exit(server$close())

  state <- attr(server, "sync")
  expect_null(state$storage_id)
})

test_that("create_document generates valid ID", {
  server <- sync_server(data_dir = tempdir())
  on.exit({
    server$close()
    unlink(file.path(tempdir(), "*.automerge"))
  })

  doc_id <- create_document(server)
  expect_type(doc_id, "character")
  expect_true(nchar(doc_id) > 20) # Base58Check encoded 16 bytes + version + checksum
  expect_true(doc_id %in% list_documents(server))
})

test_that("get_document retrieves created document", {
  server <- sync_server(data_dir = tempdir())
  on.exit({
    server$close()
    unlink(file.path(tempdir(), "*.automerge"))
  })

  doc_id <- create_document(server)
  doc <- get_document(server, doc_id)
  expect_true(inherits(doc, "am_doc"))
})

test_that("list_documents returns all documents", {
  server <- sync_server(data_dir = tempdir())
  on.exit({
    server$close()
    unlink(file.path(tempdir(), "*.automerge"))
  })

  expect_length(list_documents(server), 0)

  id1 <- create_document(server)
  id2 <- create_document(server)

  docs <- list_documents(server)
  expect_length(docs, 2)
  expect_true(id1 %in% docs)
  expect_true(id2 %in% docs)
})

test_that("print.autosync_server works", {
  server <- sync_server(data_dir = tempdir())
  on.exit(server$close())

  output <- capture.output(print(server))
  expect_true(any(grepl("Automerge Sync Server", output)))
  expect_true(any(grepl(paste0("URL: ", server$url), output, fixed = TRUE)))
})

test_that("generate_document_id creates valid IDs", {
  id <- generate_document_id()
  expect_type(id, "character")
  expect_true(nchar(id) > 20)
  # Should be decodable to 16 bytes
  bytes <- secretbase::base58dec(id, convert = FALSE)
  expect_length(bytes, 16L)
})

test_that("sync_server with custom storage_id", {
  server <- sync_server(
    storage_id = "customStorageId",
    data_dir = tempdir()
  )
  on.exit(server$close())

  state <- attr(server, "sync")
  expect_equal(state$storage_id, "customStorageId")
})

# Authentication module for autosync - OIDC JWT verification

# Module-level JWKS cache keyed by issuer URL
oidc_cache <- new.env(parent = emptyenv())

# Wrapper for interactive() to allow mocking in tests
is_interactive <- function() interactive()

#' Default OIDC issuer URL
#'
#' Returns the `OIDC_ISSUER` environment variable if set and non-empty,
#' otherwise falls back to Google (`"https://accounts.google.com"`).
#'
#' @return Character string, the OIDC issuer URL.
#'
#' @keywords internal
oidc_issuer <- function() {
  issuer <- Sys.getenv("OIDC_ISSUER")
  if (nzchar(issuer)) issuer else "https://accounts.google.com"
}

#' Discover the JWKS URI for an OIDC issuer
#'
#' @param issuer The OIDC issuer URL.
#'
#' @return Character string, the JWKS URI.
#'
#' @keywords internal
discover_jwks_uri <- function(issuer) {
  config_url <- paste0(issuer, "/.well-known/openid-configuration")
  resp <- ncurl(config_url, timeout = 5000L)

  if (is_error_value(resp$data) || resp$status != 200L) {
    stop("Failed to fetch OIDC configuration from: ", config_url)
  }

  config <- jsondec(resp$data)
  jwks_uri <- config$jwks_uri

  if (is.null(jwks_uri)) {
    stop("No jwks_uri found in OIDC configuration from: ", config_url)
  }

  jwks_uri
}

#' Fetch and parse a JWKS endpoint
#'
#' @param jwks_uri URL of the JWKS endpoint.
#'
#' @return List with `keys` (named list of jose key objects by kid) and
#'   `expiry` (POSIXct when the cache should be refreshed).
#'
#' @keywords internal
fetch_jwks <- function(jwks_uri) {
  resp <- ncurl(
    jwks_uri,
    timeout = 5000L,
    response = "Cache-Control"
  )

  if (is_error_value(resp$data) || resp$status != 200L) {
    stop("Failed to fetch JWKS from: ", jwks_uri)
  }

  jwks <- jsondec(resp$data)

  if (is.null(jwks$keys) || !length(jwks$keys)) {
    stop("No keys found in JWKS from: ", jwks_uri)
  }

  keys <- list()
  for (jwk in jwks$keys) {
    kid <- jwk$kid
    if (is.null(kid)) {
      next
    }
    key <- tryCatch(
      read_jwk(jsonenc(jwk)),
      error = function(e) NULL
    )
    if (!is.null(key)) {
      keys[[kid]] <- key
    }
  }

  # Parse Cache-Control max-age for TTL (default 1 hour)
  ttl <- 3600L
  cache_control <- resp$headers[["Cache-Control"]]
  if (!is.null(cache_control)) {
    m <- regmatches(cache_control, regexpr("max-age=(\\d+)", cache_control))
    if (length(m) && nzchar(m)) {
      ttl <- as.integer(sub("max-age=", "", m))
    }
  }

  list(keys = keys, expiry = Sys.time() + ttl)
}

#' Get a signing key for JWT verification
#'
#' Looks up the public key for the given issuer and key ID, refreshing the
#' cache if needed.
#'
#' @param issuer The OIDC issuer URL.
#' @param kid The key ID from the JWT header.
#'
#' @return A public key object, or NULL if not found.
#'
#' @keywords internal
get_signing_key <- function(issuer, kid) {
  cache_key <- issuer

  cached <- oidc_cache[[cache_key]]

  # Return cached key if available and not expired
  if (!is.null(cached) && Sys.time() < cached$expiry) {
    key <- cached$keys[[kid]]
    if (!is.null(key)) return(key)
  }

  # Refresh: discover JWKS URI (use cached URI if available)
  jwks_uri <- if (!is.null(cached$jwks_uri)) {
    cached$jwks_uri
  } else {
    discover_jwks_uri(issuer)
  }

  jwks_data <- fetch_jwks(jwks_uri)
  oidc_cache[[cache_key]] <- list(
    jwks_uri = jwks_uri,
    keys = jwks_data$keys,
    expiry = jwks_data$expiry
  )

  jwks_data$keys[[kid]]
}

#' Validate a JWT token
#'
#' Verifies a JWT locally using the issuer's public keys (JWKS) and validates
#' standard OIDC claims.
#'
#' @param token Character string, the JWT.
#' @param issuer Expected issuer URL.
#' @param client_id Expected audience (client ID).
#' @param allowed_emails Character vector of allowed email addresses (optional).
#' @param allowed_domains Character vector of allowed email domains (optional).
#' @param custom_validator Function(claims) returning TRUE/FALSE (optional).
#'
#' @return List with `valid` (logical), `email` (character or NULL),
#'   `error` (character or NULL).
#'
#' @keywords internal
validate_token <- function(
  token,
  issuer,
  client_id,
  allowed_emails = NULL,
  allowed_domains = NULL,
  custom_validator = NULL
) {
  # Extract kid from JWT header
  parts <- strsplit(token, ".", fixed = TRUE)[[1L]]
  if (length(parts) != 3L) {
    return(list(valid = FALSE, email = NULL, error = "Invalid JWT format"))
  }

  header <- tryCatch(
    jsondec(base64dec(parts[1L], url = TRUE)),
    error = function(e) NULL
  )
  if (is.null(header) || is.null(header$kid)) {
    return(list(valid = FALSE, email = NULL, error = "Invalid JWT header"))
  }

  # Get the signing key
  key <- tryCatch(
    get_signing_key(issuer, header$kid),
    error = function(e) NULL
  )
  if (is.null(key)) {
    return(list(
      valid = FALSE,
      email = NULL,
      error = "Unable to verify token signature"
    ))
  }

  # Verify signature and decode claims
  claims <- tryCatch(
    jwt_decode_sig(token, key),
    error = function(e) {
      msg <- conditionMessage(e)
      if (grepl("expired", msg, ignore.case = TRUE)) {
        structure(msg, class = "token_expired")
      } else {
        NULL
      }
    }
  )

  if (is.null(claims)) {
    return(list(
      valid = FALSE,
      email = NULL,
      error = "Token signature verification failed"
    ))
  }

  if (inherits(claims, "token_expired")) {
    return(list(valid = FALSE, email = NULL, error = "Token expired"))
  }

  # Validate issuer
  if (is.null(claims[["iss"]]) || claims[["iss"]] != issuer) {
    return(list(valid = FALSE, email = NULL, error = "Invalid token issuer"))
  }

  # Validate audience
  aud <- claims[["aud"]]
  if (is.null(aud) || !client_id %in% aud) {
    return(list(valid = FALSE, email = NULL, error = "Invalid token audience"))
  }

  # Validate expiry
  now <- as.numeric(Sys.time())
  clock_skew <- 30
  if (
    is.null(claims[["exp"]]) || as.numeric(claims[["exp"]]) < now - clock_skew
  ) {
    return(list(valid = FALSE, email = NULL, error = "Token expired"))
  }

  # Validate issued-at
  if (
    !is.null(claims[["iat"]]) && as.numeric(claims[["iat"]]) > now + clock_skew
  ) {
    return(list(
      valid = FALSE,
      email = NULL,
      error = "Token issued in the future"
    ))
  }

  email <- claims[["email"]]

  # Check email_verified if present
  if (isFALSE(claims[["email_verified"]])) {
    return(list(
      valid = FALSE,
      email = email,
      error = "Email not verified"
    ))
  }

  # An allowlist requires a present email and email_verified == TRUE.
  if (!is.null(allowed_emails) || !is.null(allowed_domains)) {
    if (is.null(email) || !nzchar(email)) {
      return(list(
        valid = FALSE,
        email = NULL,
        error = "Token has no email claim"
      ))
    }
    if (!isTRUE(claims[["email_verified"]])) {
      return(list(
        valid = FALSE,
        email = email,
        error = "Email not verified"
      ))
    }
  }

  if (!is.null(allowed_emails)) {
    if (!email %in% allowed_emails) {
      return(list(
        valid = FALSE,
        email = email,
        error = "Email not in allowlist"
      ))
    }
  }

  if (!is.null(allowed_domains)) {
    domain <- sub(".*@", "", email)
    if (!domain %in% allowed_domains) {
      return(list(valid = FALSE, email = email, error = "Domain not allowed"))
    }
  }

  if (!is.null(custom_validator)) {
    if (!isTRUE(custom_validator(claims))) {
      return(list(
        valid = FALSE,
        email = email,
        error = "Custom validation failed"
      ))
    }
  }

  list(valid = TRUE, email = email, error = NULL)
}

#' Create an authentication configuration
#'
#' Creates a configuration object for enabling OIDC JWT authentication
#' on an autosync server. When enabled, clients must include a valid JWT
#' (ID token) as a Bearer token in the Authorization header of the WebSocket
#' upgrade request. Connections without valid credentials are rejected
#' immediately at connection time.
#'
#' Works with any OIDC-compliant identity provider: Google, Microsoft Entra,
#' Okta, Auth0, etc.
#'
#' @param client_id The OIDC client ID (application ID). Validated against the
#'   `aud` claim in JWTs. Defaults to the `OIDC_CLIENT_ID` environment variable.
#' @param issuer The OIDC issuer URL. This is used to discover the provider's
#'   public keys via the `.well-known/openid-configuration` endpoint, and to
#'   validate the `iss` claim in JWTs. Defaults to the `OIDC_ISSUER`
#'   environment variable, falling back to Google
#'   (`"https://accounts.google.com"`).
#' @param allowed_emails Character vector of allowed email addresses. When set,
#'   a token is rejected unless it carries an `email` claim with
#'   `email_verified` explicitly `TRUE`.
#' @param allowed_domains Character vector of allowed email domains
#'   (e.g., "mycompany.com"). Subject to the same verified-email requirement as
#'   `allowed_emails`.
#' @param custom_validator Function(claims) returning TRUE/FALSE for
#'   custom validation logic. Receives the decoded JWT claims as a list.
#'
#' @return An object of class `"autosync_auth_config"`.
#'
#' @examples
#' # Google (default issuer)
#' auth_config(
#'   client_id = "123456789.apps.googleusercontent.com",
#'   allowed_domains = "mycompany.com"
#' )
#'
#' # Microsoft Entra
#' auth_config(
#'   client_id = "abcdef-1234-5678",
#'   issuer = "https://login.microsoftonline.com/common/v2.0",
#'   allowed_emails = "alice@mycompany.com"
#' )
#'
#' # Custom validator
#' auth_config(
#'   client_id = "0oaXXXXXXXX",
#'   issuer = "https://dev-123456.okta.com/oauth2/default",
#'   custom_validator = function(claims) "editors" %in% claims$groups
#' )
#'
#' @export
auth_config <- function(
  client_id = Sys.getenv("OIDC_CLIENT_ID"),
  issuer = oidc_issuer(),
  allowed_emails = NULL,
  allowed_domains = NULL,
  custom_validator = NULL
) {
  if (!is.character(issuer) || length(issuer) != 1L) {
    stop("'issuer' must be a single character string (OIDC issuer URL)")
  }

  if (
    !is.character(client_id) || length(client_id) != 1L || !nzchar(client_id)
  ) {
    stop(
      "'client_id' must be set (or set the OIDC_CLIENT_ID environment variable)"
    )
  }

  structure(
    list(
      issuer = issuer,
      client_id = client_id,
      allowed_emails = allowed_emails,
      allowed_domains = allowed_domains,
      custom_validator = custom_validator
    ),
    class = "autosync_auth_config"
  )
}

#' Obtain an OIDC token interactively
#'
#' Performs the OAuth 2.0 Authorization Code flow with PKCE to obtain a JWT
#' (ID token) from an OIDC provider, delegating the browser handshake and token
#' exchange to \pkg{httr2}. Endpoints are discovered from the issuer's
#' `.well-known` metadata via [httr2::oauth_server_metadata()], and the flow is
#' run by [httr2::oauth_flow_auth_code()]: it opens the system browser, listens
#' on a loopback redirect for the callback, and returns the ID token for use
#' with [sync_fetch()].
#'
#' For Google, register the OAuth client as a "Desktop app" and set both
#' `OIDC_CLIENT_ID` and `OIDC_CLIENT_SECRET`. Google's Desktop app secret is
#' required in the token exchange but, unlike a "Web application" secret, is
#' not treated as confidential: Google states that for installed apps "the
#' client secret is obviously not treated as a secret"
#' (<https://developers.google.com/identity/protocols/oauth2#installed>),
#' consistent with the OAuth 2.0 for Native Apps standard
#' (RFC 8252 section 8.5,
#' <https://datatracker.ietf.org/doc/html/rfc8252#section-8.5>).
#' Providers that support native / public clients (Microsoft Entra, Okta,
#' Auth0, etc.) need only `client_id`, authenticating via PKCE alone; leave
#' `client_secret` unset for these.
#'
#' @param client_id The OIDC client ID (application ID). Defaults to the
#'   `OIDC_CLIENT_ID` environment variable.
#' @param client_secret The OIDC client secret. Required by Google (Desktop
#'   app) and "Web application" client types; leave unset for native / public
#'   clients, which authenticate via PKCE alone. Defaults to the
#'   `OIDC_CLIENT_SECRET` environment variable.
#' @param issuer The OIDC issuer URL. Defaults to the `OIDC_ISSUER`
#'   environment variable, falling back to Google
#'   (`"https://accounts.google.com"`).
#' @param scopes Space-separated OAuth scopes to request. Default
#'   `"openid email"`.
#' @param redirect_uri Local redirect URI for the OAuth callback. Defaults to
#'   [httr2::oauth_redirect_uri()], i.e. `"http://localhost"` with an
#'   OS-assigned random port (or the `HTTR2_OAUTH_REDIRECT_URL` environment
#'   variable on hosted platforms). Supply an explicit port (e.g.
#'   `"http://localhost:8080"`) when your OIDC provider requires the redirect
#'   URI to match a pre-registered value.
#'
#' @return A JWT (ID token) as a character string.
#'
#' @examplesIf interactive()
#' # Uses OIDC_CLIENT_ID and OIDC_CLIENT_SECRET env vars by default
#' token <- sync_token()
#'
#' # Or supply credentials directly
#' token <- sync_token(
#'   client_id = "YOUR_CLIENT_ID.apps.googleusercontent.com",
#'   client_secret = "YOUR_CLIENT_SECRET"
#' )
#'
#' # Use with sync_fetch
#' doc <- sync_fetch(server$url, "myDocId", token = token, tls = tls)
#'
#' @export
sync_token <- function(
  client_id = Sys.getenv("OIDC_CLIENT_ID"),
  client_secret = Sys.getenv("OIDC_CLIENT_SECRET"),
  issuer = oidc_issuer(),
  scopes = "openid email",
  redirect_uri = oauth_redirect_uri()
) {
  if (!is_interactive()) {
    stop("sync_token() requires an interactive session")
  }

  if (
    !is.character(client_id) || length(client_id) != 1L || !nzchar(client_id)
  ) {
    stop(
      "'client_id' must be set (or set the OIDC_CLIENT_ID environment variable)"
    )
  }

  # Discover the provider's OAuth/OIDC endpoints from its well-known metadata.
  metadata <- oauth_server_metadata(issuer)
  if (
    is.null(metadata$authorization_endpoint) ||
      is.null(metadata$token_endpoint)
  ) {
    stop("OIDC configuration missing authorization_endpoint or token_endpoint")
  }

  # An empty secret means a public / native client: authenticate via PKCE alone.
  client <- oauth_client(
    id = client_id,
    secret = if (nzchar(client_secret)) client_secret else NULL,
    metadata = metadata
  )

  # Authorization Code + PKCE: opens the browser, listens on the loopback
  # redirect for the callback, then exchanges the code for tokens.
  token <- oauth_flow_auth_code(
    client,
    scope = scopes,
    redirect_uri = redirect_uri
  )

  if (is.null(token$id_token)) {
    stop("No ID token in response (ensure 'openid' scope is requested)")
  }

  token$id_token
}

#' Authenticate a client from HTTP request headers
#'
#' Extracts and validates a Bearer token (JWT) from the Authorization header
#' of the WebSocket upgrade request.
#'
#' @param auth_config A `autosync_auth_config` object.
#' @param headers Named list of HTTP request headers.
#'
#' @return List with `valid` (logical), `email` (character or NULL),
#'   `error` (character or NULL).
#'
#' @keywords internal
authenticate_header <- function(auth_config, headers) {
  auth_fail <- list(
    valid = FALSE,
    email = NULL,
    error = "Authentication failed"
  )

  if (
    is.null(headers) || !length(headers) || !"Authorization" %in% names(headers)
  ) {
    return(auth_fail)
  }

  auth_header <- unname(headers["Authorization"])
  if (!startsWith(auth_header, "Bearer ")) {
    return(auth_fail)
  }

  token <- substring(auth_header, 8L)

  # JWT format: Base64URL chars and dots, typically 800-2000 chars
  if (
    nchar(token) < 20L ||
      nchar(token) > 8192L ||
      grepl("[^A-Za-z0-9_.-]", token)
  ) {
    return(auth_fail)
  }

  validate_token(
    token = token,
    issuer = auth_config$issuer,
    client_id = auth_config$client_id,
    allowed_emails = auth_config$allowed_emails,
    allowed_domains = auth_config$allowed_domains,
    custom_validator = auth_config$custom_validator
  )
}

# Tests for OIDC JWT authentication module

test_that("is_interactive reflects the running session's interactivity", {
  # The wrapper exists so it can be mocked elsewhere; unmocked it just defers
  # to base::interactive().
  expect_identical(autosync:::is_interactive(), interactive())
})

# Helper: create a test JWT signed with a given RSA key
create_test_jwt <- function(claims, key, kid = "test-kid-1") {
  claim <- do.call(jose::jwt_claim, claims)
  jose::jwt_encode_sig(claim, key, header = list(kid = kid))
}

# Helper: standard valid claims
valid_claims <- function(
  issuer = "https://accounts.google.com",
  client_id = "test-client-id",
  email = "user@test.com"
) {
  now <- as.integer(Sys.time())
  list(
    iss = issuer,
    aud = client_id,
    sub = "1234567890",
    email = email,
    email_verified = TRUE,
    iat = now - 60L,
    exp = now + 3600L
  )
}

# ---- auth_config tests ----

test_that("auth_config creates valid configuration", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_domains = "example.com",
    allowed_emails = "test@example.com"
  )

  expect_s3_class(cfg, "autosync_auth_config")
  expect_equal(cfg$issuer, "https://accounts.google.com")
  expect_equal(cfg$client_id, "test-client-id")
  expect_equal(cfg$allowed_domains, "example.com")
  expect_equal(cfg$allowed_emails, "test@example.com")
})

test_that("oidc_issuer reads OIDC_ISSUER env var", {
  old <- Sys.getenv("OIDC_ISSUER")
  on.exit(if (nzchar(old)) Sys.setenv(OIDC_ISSUER = old) else Sys.unsetenv("OIDC_ISSUER"))
  Sys.setenv(OIDC_ISSUER = "https://login.microsoftonline.com/common/v2.0")
  expect_equal(oidc_issuer(), "https://login.microsoftonline.com/common/v2.0")
})

test_that("oidc_issuer falls back to Google when OIDC_ISSUER is unset", {
  old <- Sys.getenv("OIDC_ISSUER")
  on.exit(if (nzchar(old)) Sys.setenv(OIDC_ISSUER = old) else Sys.unsetenv("OIDC_ISSUER"))
  Sys.unsetenv("OIDC_ISSUER")
  expect_equal(oidc_issuer(), "https://accounts.google.com")
})

test_that("oidc_issuer falls back to Google when OIDC_ISSUER is empty", {
  old <- Sys.getenv("OIDC_ISSUER")
  on.exit(if (nzchar(old)) Sys.setenv(OIDC_ISSUER = old) else Sys.unsetenv("OIDC_ISSUER"))
  Sys.setenv(OIDC_ISSUER = "")
  expect_equal(oidc_issuer(), "https://accounts.google.com")
})

test_that("auth_config defaults issuer from OIDC_ISSUER env var", {
  old <- Sys.getenv("OIDC_ISSUER")
  on.exit(if (nzchar(old)) Sys.setenv(OIDC_ISSUER = old) else Sys.unsetenv("OIDC_ISSUER"))
  Sys.setenv(OIDC_ISSUER = "https://login.microsoftonline.com/common/v2.0")
  cfg <- auth_config(client_id = "test-id")
  expect_equal(cfg$issuer, "https://login.microsoftonline.com/common/v2.0")
})

test_that("auth_config validates issuer and client_id", {
  expect_snapshot(auth_config(issuer = 123, client_id = "x"), error = TRUE)
  expect_snapshot(auth_config(issuer = "x", client_id = 123), error = TRUE)
  expect_snapshot(auth_config(client_id = ""), error = TRUE)
})

test_that("auth_config has NULL defaults for optional parameters", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  expect_null(cfg$allowed_emails)
  expect_null(cfg$allowed_domains)
  expect_null(cfg$custom_validator)
})

test_that("auth_config stores custom_validator function", {
  validator <- function(claims) "editors" %in% claims$groups
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id",
    custom_validator = validator
  )
  expect_identical(cfg$custom_validator, validator)
})

# ---- authenticate_header tests ----

test_that("authenticate_header returns error for NULL headers", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  result <- authenticate_header(cfg, NULL)
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")
})

test_that("authenticate_header returns generic error for invalid credentials", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )

  # Missing headers
  result <- authenticate_header(cfg, c())
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")

  # Missing Authorization header
  result <- authenticate_header(cfg, c(Host = "localhost"))
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")

  # Wrong scheme (not Bearer)
  result <- authenticate_header(cfg, c(Authorization = "Basic dXNlcjpwYXNz"))
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")

  # Too short token
  result <- authenticate_header(cfg, c(Authorization = "Bearer short"))
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")

  # Too long (over 8192 chars)
  result <- authenticate_header(
    cfg,
    c(Authorization = paste("Bearer", strrep("a", 9000)))
  )
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")

  # Contains special chars invalid for JWT
  result <- authenticate_header(
    cfg,
    c(Authorization = paste0("Bearer ", strrep("a", 25), "$%^&", strrep("b", 25)))
  )
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")

  # Empty token after Bearer
  result <- authenticate_header(cfg, c(Authorization = "Bearer "))
  expect_false(result$valid)
  expect_equal(result$error, "Authentication failed")
})

test_that("authenticate_header accepts token at minimum length boundary", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  local_mocked_bindings(
    validate_token = function(...) list(valid = TRUE, email = "a@b.com", error = NULL)
  )
  result <- authenticate_header(cfg, c(Authorization = paste("Bearer", strrep("a", 20))))
  expect_true(result$valid)
})

test_that("authenticate_header accepts token at maximum length boundary", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  local_mocked_bindings(
    validate_token = function(...) list(valid = TRUE, email = "a@b.com", error = NULL)
  )
  result <- authenticate_header(cfg, c(Authorization = paste("Bearer", strrep("a", 8192))))
  expect_true(result$valid)
})

test_that("authenticate_header accepts valid JWT characters", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  local_mocked_bindings(
    validate_token = function(...) list(valid = TRUE, email = "a@b.com", error = NULL)
  )
  # All allowed characters: A-Za-z0-9_.-
  token <- "ABCDEFghij0123456789_.-"
  result <- authenticate_header(cfg, c(Authorization = paste("Bearer", token)))
  expect_true(result$valid)
})

test_that("authenticate_header passes config to validate_token", {
  cfg <- auth_config(
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_emails = c("user@test.com"),
    allowed_domains = c("test.com"),
    custom_validator = function(x) TRUE
  )

  captured_args <- NULL
  local_mocked_bindings(
    validate_token = function(...) {
      captured_args <<- list(...)
      list(valid = TRUE, email = "user@test.com", error = NULL)
    }
  )

  token <- strrep("x", 30)
  authenticate_header(cfg, c(Authorization = paste("Bearer", token)))

  expect_equal(captured_args$token, token)
  expect_equal(captured_args$issuer, "https://accounts.google.com")
  expect_equal(captured_args$client_id, "test-client-id")
  expect_equal(captured_args$allowed_emails, c("user@test.com"))
  expect_equal(captured_args$allowed_domains, c("test.com"))
  expect_true(is.function(captured_args$custom_validator))
})

# ---- validate_token JWT tests ----

test_that("validate_token verifies a valid JWT", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()

  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )

  expect_true(result$valid)
  expect_equal(result$email, "user@test.com")
  expect_null(result$error)
})

test_that("validate_token rejects invalid JWT format", {
  result <- validate_token(
    token = "not-a-jwt",
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Invalid JWT format")
})

test_that("validate_token rejects JWT with bad header", {
  result <- validate_token(
    token = "!!!.bbb.ccc",
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Invalid JWT header")
})

test_that("validate_token rejects JWT with unknown kid", {
  key <- openssl::rsa_keygen(2048)
  claims <- valid_claims()
  jwt <- create_test_jwt(claims, key, kid = "unknown-kid")

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) NULL
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Unable to verify token signature")
})

test_that("validate_token rejects JWT signed with wrong key", {
  signing_key <- openssl::rsa_keygen(2048)
  wrong_key <- openssl::rsa_keygen(2048)
  wrong_pubkey <- as.list(wrong_key)$pubkey
  claims <- valid_claims()

  jwt <- create_test_jwt(claims, signing_key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) wrong_pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Token signature verification failed")
})

test_that("validate_token rejects wrong issuer", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(issuer = "https://evil.com")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Invalid token issuer")
})

test_that("validate_token rejects wrong audience", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(client_id = "wrong-client-id")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Invalid token audience")
})

test_that("validate_token rejects expired JWT", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$exp <- as.integer(Sys.time()) - 120L  # expired 2 minutes ago
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Token expired")
})

test_that("validate_token rejects JWT issued in the future", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$iat <- as.integer(Sys.time()) + 120L  # 2 minutes in future
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Token issued in the future")
})

test_that("validate_token rejects unverified email", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$email_verified <- FALSE
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Email not verified")
})

test_that("validate_token rejects email not in allowlist", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(email = "stranger@other.com")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_emails = c("alice@test.com", "bob@test.com")
  )
  expect_false(result$valid)
  expect_equal(result$email, "stranger@other.com")
  expect_equal(result$error, "Email not in allowlist")
})

test_that("validate_token rejects missing email claim when allowlist is set", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$email <- NULL
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result_emails <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_emails = c("alice@test.com")
  )
  expect_false(result_emails$valid)
  expect_equal(result_emails$error, "Token has no email claim")

  result_domains <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_domains = c("test.com")
  )
  expect_false(result_domains$valid)
  expect_equal(result_domains$error, "Token has no email claim")
})

test_that("validate_token requires verified email when allowlist is set", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  # email_verified absent (not explicitly TRUE)
  claims <- valid_claims(email = "alice@test.com")
  claims$email_verified <- NULL
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_emails = c("alice@test.com")
  )
  expect_false(result$valid)
  expect_equal(result$error, "Email not verified")
})

test_that("validate_token allows unverified email when no allowlist is set", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  # email_verified absent, no allowlist -> not gated on email, so accepted
  claims <- valid_claims(email = "alice@test.com")
  claims$email_verified <- NULL
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_true(result$valid)
  expect_equal(result$email, "alice@test.com")
})

test_that("validate_token accepts email in allowlist", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(email = "alice@test.com")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_emails = c("alice@test.com", "bob@test.com")
  )
  expect_true(result$valid)
  expect_equal(result$email, "alice@test.com")
})

test_that("validate_token rejects disallowed domain", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(email = "user@evil.com")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_domains = c("trusted.com", "partner.org")
  )
  expect_false(result$valid)
  expect_equal(result$email, "user@evil.com")
  expect_equal(result$error, "Domain not allowed")
})

test_that("validate_token accepts allowed domain", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(email = "user@trusted.com")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_domains = c("trusted.com", "partner.org")
  )
  expect_true(result$valid)
  expect_equal(result$email, "user@trusted.com")
})

test_that("validate_token rejects when custom_validator returns FALSE", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$groups <- "viewers"
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    custom_validator = function(claims) "editors" %in% claims$groups
  )
  expect_false(result$valid)
  expect_equal(result$email, "user@test.com")
  expect_equal(result$error, "Custom validation failed")
})

test_that("validate_token accepts when custom_validator returns TRUE", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$groups <- "editors"
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    custom_validator = function(claims) "editors" %in% claims$groups
  )
  expect_true(result$valid)
  expect_equal(result$email, "user@test.com")
})

test_that("validate_token applies email, domain, and custom checks in order", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims(email = "user@trusted.com")
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id",
    allowed_emails = c("user@trusted.com"),
    allowed_domains = c("trusted.com"),
    custom_validator = function(claims) TRUE
  )
  expect_true(result$valid)
  expect_equal(result$email, "user@trusted.com")
})

# ---- OIDC discovery and JWKS cache tests ----

test_that("discover_jwks_uri fetches from well-known endpoint", {
  local_mocked_bindings(
    ncurl = function(url, ...) {
      expect_match(url, "\\.well-known/openid-configuration$")
      list(
        data = charToRaw('{"jwks_uri":"https://www.googleapis.com/oauth2/v3/certs"}'),
        status = 200L
      )
    },
    is_error_value = function(x) FALSE
  )

  uri <- discover_jwks_uri("https://accounts.google.com")
  expect_equal(uri, "https://www.googleapis.com/oauth2/v3/certs")
})

test_that("discover_jwks_uri errors on failed request", {
  local_mocked_bindings(
    ncurl = function(...) list(
      data = structure(5L, class = "errorValue"),
      status = 0L
    ),
    is_error_value = function(x) inherits(x, "errorValue")
  )

  expect_error(
    discover_jwks_uri("https://bad-issuer.com"),
    "Failed to fetch OIDC configuration"
  )
})

test_that("discover_jwks_uri errors on non-200 status", {
  local_mocked_bindings(
    ncurl = function(...) list(
      data = charToRaw("Not Found"),
      status = 404L
    ),
    is_error_value = function(x) FALSE
  )

  expect_error(
    discover_jwks_uri("https://bad-issuer.com"),
    "Failed to fetch OIDC configuration"
  )
})

test_that("discover_jwks_uri errors on missing jwks_uri", {
  local_mocked_bindings(
    ncurl = function(...) list(
      data = charToRaw('{"issuer":"https://accounts.google.com"}'),
      status = 200L
    ),
    is_error_value = function(x) FALSE
  )

  expect_error(
    discover_jwks_uri("https://accounts.google.com"),
    "No jwks_uri found"
  )
})

test_that("get_signing_key caches and refreshes keys", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  jwk_json <- jose::write_jwk(pubkey)
  jwk <- jsondec(jwk_json)
  jwk$kid <- "cached-kid"

  # Clear cache
  rm(list = ls(oidc_cache), envir = oidc_cache)

  fetch_count <- 0L
  local_mocked_bindings(
    discover_jwks_uri = function(issuer) "https://example.com/jwks",
    fetch_jwks = function(jwks_uri) {
      fetch_count <<- fetch_count + 1L
      list(
        keys = list("cached-kid" = pubkey),
        expiry = Sys.time() + 3600
      )
    }
  )

  # First call should fetch
  result <- get_signing_key("https://example.com", "cached-kid")
  expect_false(is.null(result))
  expect_equal(fetch_count, 1L)

  # Second call should use cache
  result2 <- get_signing_key("https://example.com", "cached-kid")
  expect_false(is.null(result2))
  expect_equal(fetch_count, 1L)

  # Clean up
  rm(list = ls(oidc_cache), envir = oidc_cache)
})

test_that("get_signing_key refreshes on unknown kid", {
  key1 <- openssl::rsa_keygen(2048)
  pubkey1 <- as.list(key1)$pubkey
  key2 <- openssl::rsa_keygen(2048)
  pubkey2 <- as.list(key2)$pubkey

  # Clear cache
  rm(list = ls(oidc_cache), envir = oidc_cache)

  call_count <- 0L
  local_mocked_bindings(
    discover_jwks_uri = function(issuer) "https://example.com/jwks",
    fetch_jwks = function(jwks_uri) {
      call_count <<- call_count + 1L
      keys <- list("kid-1" = pubkey1)
      if (call_count >= 2L) keys[["kid-2"]] <- pubkey2
      list(keys = keys, expiry = Sys.time() + 3600)
    }
  )

  # First fetch gets kid-1
  result <- get_signing_key("https://example.com", "kid-1")
  expect_false(is.null(result))
  expect_equal(call_count, 1L)

  # Unknown kid-2 triggers refresh
  result2 <- get_signing_key("https://example.com", "kid-2")
  expect_false(is.null(result2))
  expect_equal(call_count, 2L)

  # Clean up
  rm(list = ls(oidc_cache), envir = oidc_cache)
})

test_that("get_signing_key returns NULL for unknown kid after refresh", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey

  # Clear cache
  rm(list = ls(oidc_cache), envir = oidc_cache)

  local_mocked_bindings(
    discover_jwks_uri = function(issuer) "https://example.com/jwks",
    fetch_jwks = function(jwks_uri) {
      list(
        keys = list("kid-1" = pubkey),
        expiry = Sys.time() + 3600
      )
    }
  )

  # Request a kid that doesn't exist even after fetch
  result <- get_signing_key("https://example.com", "nonexistent-kid")
  expect_null(result)

  # Clean up
  rm(list = ls(oidc_cache), envir = oidc_cache)
})

test_that("get_signing_key refreshes expired cache", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey

  # Clear cache
  rm(list = ls(oidc_cache), envir = oidc_cache)

  # Seed cache with expired entry
  oidc_cache[["https://example.com"]] <- list(
    jwks_uri = "https://example.com/jwks",
    keys = list("kid-1" = pubkey),
    expiry = Sys.time() - 1  # already expired
  )

  fetch_count <- 0L
  local_mocked_bindings(
    fetch_jwks = function(jwks_uri) {
      fetch_count <<- fetch_count + 1L
      list(keys = list("kid-1" = pubkey), expiry = Sys.time() + 3600)
    }
  )

  result <- get_signing_key("https://example.com", "kid-1")
  expect_false(is.null(result))
  expect_equal(fetch_count, 1L)

  # Clean up
  rm(list = ls(oidc_cache), envir = oidc_cache)
})

# ---- fetch_jwks tests ----

test_that("fetch_jwks parses keys and cache-control", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  jwk <- jsondec(jose::write_jwk(pubkey))
  jwk$kid <- "test-kid"
  jwks_json <- jsonenc(list(keys = list(jwk)))

  local_mocked_bindings(
    ncurl = function(url, ...) list(
      data = charToRaw(jwks_json),
      status = 200L,
      headers = list("Cache-Control" = "public, max-age=7200")
    ),
    is_error_value = function(x) FALSE
  )

  result <- fetch_jwks("https://example.com/jwks")
  expect_true("test-kid" %in% names(result$keys))
  expect_true(result$expiry > Sys.time() + 7000)
})

test_that("fetch_jwks uses default TTL without cache-control", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  jwk <- jsondec(jose::write_jwk(pubkey))
  jwk$kid <- "test-kid"
  jwks_json <- jsonenc(list(keys = list(jwk)))

  local_mocked_bindings(
    ncurl = function(url, ...) list(
      data = charToRaw(jwks_json),
      status = 200L,
      headers = list()
    ),
    is_error_value = function(x) FALSE
  )

  result <- fetch_jwks("https://example.com/jwks")
  expect_true("test-kid" %in% names(result$keys))
  expect_true(result$expiry > Sys.time() + 3500)
  expect_true(result$expiry < Sys.time() + 3700)
})

test_that("fetch_jwks errors on failed request", {
  local_mocked_bindings(
    ncurl = function(...) list(
      data = structure(5L, class = "errorValue"),
      status = 0L
    ),
    is_error_value = function(x) inherits(x, "errorValue")
  )

  expect_error(fetch_jwks("https://example.com/jwks"), "Failed to fetch JWKS")
})

test_that("fetch_jwks errors on non-200 status", {
  local_mocked_bindings(
    ncurl = function(...) list(
      data = charToRaw("Server Error"),
      status = 500L
    ),
    is_error_value = function(x) FALSE
  )

  expect_error(fetch_jwks("https://example.com/jwks"), "Failed to fetch JWKS")
})

test_that("fetch_jwks errors on empty keys", {
  local_mocked_bindings(
    ncurl = function(...) list(
      data = charToRaw('{"keys":[]}'),
      status = 200L,
      headers = list()
    ),
    is_error_value = function(x) FALSE
  )

  expect_error(fetch_jwks("https://example.com/jwks"), "No keys found")
})

test_that("fetch_jwks skips keys without kid", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  jwk_with_kid <- jsondec(jose::write_jwk(pubkey))
  jwk_with_kid$kid <- "good-kid"
  jwk_no_kid <- jsondec(jose::write_jwk(pubkey))
  jwks_json <- jsonenc(list(keys = list(jwk_no_kid, jwk_with_kid)))

  local_mocked_bindings(
    ncurl = function(...) list(
      data = charToRaw(jwks_json),
      status = 200L,
      headers = list()
    ),
    is_error_value = function(x) FALSE
  )

  result <- fetch_jwks("https://example.com/jwks")
  expect_equal(names(result$keys), "good-kid")
})

test_that("fetch_jwks skips unparseable keys", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  good_jwk <- jsondec(jose::write_jwk(pubkey))
  good_jwk$kid <- "good-kid"
  bad_jwk <- list(kid = "bad-kid", kty = "INVALID")
  jwks_json <- jsonenc(list(keys = list(bad_jwk, good_jwk)))

  local_mocked_bindings(
    ncurl = function(...) list(
      data = charToRaw(jwks_json),
      status = 200L,
      headers = list()
    ),
    is_error_value = function(x) FALSE
  )

  result <- fetch_jwks("https://example.com/jwks")
  expect_equal(names(result$keys), "good-kid")
})

# ---- validate_token edge cases ----

test_that("validate_token returns error when get_signing_key throws", {
  key <- openssl::rsa_keygen(2048)
  claims <- valid_claims()
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) stop("network error")
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Unable to verify token signature")
})

test_that("validate_token rejects JWT header without kid", {
  # Create a JWT-like token with a valid base64url header that has no kid
  header_json <- '{"alg":"RS256","typ":"JWT"}'
  header_b64 <- jose::base64url_encode(charToRaw(header_json))
  token <- paste(header_b64, "payload", "signature", sep = ".")

  result <- validate_token(
    token = token,
    issuer = "https://accounts.google.com",
    client_id = "test-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Invalid JWT header")
})

# ---- validate_token missing exp ----

test_that("validate_token rejects JWT with missing exp claim", {
  key <- openssl::rsa_keygen(2048)
  pubkey <- as.list(key)$pubkey
  claims <- valid_claims()
  claims$exp <- NULL
  jwt <- create_test_jwt(claims, key)

  local_mocked_bindings(
    get_signing_key = function(issuer, kid) pubkey
  )

  result <- validate_token(
    token = jwt,
    issuer = "https://accounts.google.com",
    client_id = "test-client-id"
  )
  expect_false(result$valid)
  expect_equal(result$error, "Token expired")
})

# ---- sync_token tests ----

# Stub oauth_server_metadata(): returns a metadata object advertising the
# endpoints sync_token reads. The "httr2_oauth_server_metadata" class lets the
# real oauth_client() accept it via metadata=.
stub_metadata <- function(
  authorization_endpoint = "https://auth.example.com/authorize",
  token_endpoint = "https://auth.example.com/token"
) {
  function(issuer, ...) {
    structure(
      list(
        authorization_endpoint = authorization_endpoint,
        token_endpoint = token_endpoint
      ),
      class = "httr2_oauth_server_metadata"
    )
  }
}

test_that("sync_token errors in non-interactive session", {
  local_mocked_bindings(
    is_interactive = function() FALSE
  )
  expect_error(
    sync_token(client_id = "test-id"),
    "requires an interactive session"
  )
})

test_that("sync_token errors with invalid client_id", {
  local_mocked_bindings(
    is_interactive = function() TRUE
  )
  expect_error(sync_token(client_id = ""), "'client_id' must be set")
  expect_error(sync_token(client_id = 123), "'client_id' must be set")
})

test_that("sync_token errors on metadata missing endpoints", {
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = function(issuer, ...) list(issuer = issuer)
  )

  expect_error(
    sync_token(client_id = "test-id", issuer = "https://example.com"),
    "missing authorization_endpoint or token_endpoint"
  )
})

test_that("sync_token succeeds with valid OAuth flow", {
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = stub_metadata(),
    oauth_flow_auth_code = function(client, ...) {
      httr2::oauth_token("access", id_token = "mock.jwt.token")
    }
  )

  result <- sync_token(
    client_id = "test-client",
    client_secret = "",
    issuer = "https://issuer.example.com"
  )

  expect_equal(result, "mock.jwt.token")
})

test_that("sync_token errors when response has no id_token", {
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = stub_metadata(),
    oauth_flow_auth_code = function(client, ...) httr2::oauth_token("access")
  )

  expect_error(
    sync_token(
      client_id = "test-client",
      client_secret = "",
      issuer = "https://issuer.example.com"
    ),
    "No ID token in response"
  )
})

test_that("sync_token propagates OAuth flow errors", {
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = stub_metadata(),
    oauth_flow_auth_code = function(client, ...) {
      stop("Authentication failure: state does not match")
    }
  )

  expect_error(
    sync_token(
      client_id = "test-client",
      client_secret = "",
      issuer = "https://issuer.example.com"
    ),
    "state does not match"
  )
})

test_that("sync_token builds a confidential client when a secret is given", {
  captured <- NULL
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = stub_metadata(),
    oauth_flow_auth_code = function(client, ...) {
      captured <<- client
      httr2::oauth_token("access", id_token = "jwt")
    }
  )

  sync_token(
    client_id = "test-client",
    client_secret = "my-secret",
    issuer = "https://issuer.example.com"
  )

  expect_s3_class(captured, "httr2_oauth_client")
  expect_equal(captured$id, "test-client")
  expect_equal(captured$secret, "my-secret")
  # token_url comes from the discovered metadata.
  expect_equal(captured$token_url, "https://auth.example.com/token")
})

test_that("sync_token builds a public client when no secret is given", {
  captured <- NULL
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = stub_metadata(),
    oauth_flow_auth_code = function(client, ...) {
      captured <<- client
      httr2::oauth_token("access", id_token = "jwt")
    }
  )

  sync_token(
    client_id = "test-client",
    client_secret = "",
    issuer = "https://issuer.example.com"
  )

  # Empty secret yields a public client: PKCE only, no secret.
  expect_s3_class(captured, "httr2_oauth_client")
  expect_null(captured$secret)
})

test_that("sync_token forwards endpoints, scopes, and redirect_uri", {
  captured <- NULL
  local_mocked_bindings(
    is_interactive = function() TRUE,
    oauth_server_metadata = stub_metadata(),
    oauth_flow_auth_code = function(client, scope, redirect_uri, ...) {
      captured <<- list(
        auth_url = client$auth_url,
        token_url = client$token_url,
        scope = scope,
        redirect_uri = redirect_uri
      )
      httr2::oauth_token("access", id_token = "jwt")
    }
  )

  sync_token(
    client_id = "test-client",
    client_secret = "",
    issuer = "https://issuer.example.com",
    scopes = "openid email profile",
    redirect_uri = "http://localhost:8080"
  )

  # Endpoints ride on the client (from metadata); scope/redirect_uri pass through.
  expect_equal(captured$auth_url, "https://auth.example.com/authorize")
  expect_equal(captured$token_url, "https://auth.example.com/token")
  expect_equal(captured$scope, "openid email profile")
  expect_equal(captured$redirect_uri, "http://localhost:8080")
})

# ---- server integration tests ----

test_that("server requires TLS when auth is enabled", {
  expect_snapshot(
    sync_server(
      auth = auth_config(
        issuer = "https://accounts.google.com",
        client_id = "test-id"
      )
    ),
    error = TRUE
  )
})

test_that("server allows auth with TLS configured", {
  skip_on_cran()

  cert <- nanonext::write_cert()
  tls <- nanonext::tls_config(server = cert$server)

  server <- sync_server(
    tls = tls,
    auth = auth_config(
      issuer = "https://accounts.google.com",
      client_id = "test-id",
      allowed_emails = "test@example.com"
    )
  )
  on.exit(server$close())

  expect_s3_class(server, "autosync_server")
  state <- attr(server, "sync")
  expect_s3_class(state$auth, "autosync_auth_config")
})

test_that("server rejects unauthenticated client when auth enabled", {
  skip_on_cran()

  cert <- nanonext::write_cert()
  tls <- nanonext::tls_config(server = cert$server)
  client_tls <- nanonext::tls_config(client = cert$client)

  server <- sync_server(
    tls = tls,
    auth = auth_config(
      issuer = "https://accounts.google.com",
      client_id = "test-id",
      allowed_emails = "allowed@test.com"
    )
  )
  on.exit(server$close())
  server$start()

  # Client without token should be rejected
  expect_snapshot(
    sync_fetch(
      url = server$url,
      doc_id = generate_document_id(),
      tls = client_tls
    ),
    error = TRUE
  )
})

# Document and sync state storage functions for autosync

#' Save a document to disk
#'
#' @param server An autosync_server object.
#' @param doc_id Document ID string.
#' @param doc Automerge document object.
#'
#' @noRd
save_document <- function(server, doc_id, doc) {
  path <- file.path(server$data_dir, paste0(doc_id, ".automerge"))
  writeBin(am_save(doc), path)
}

#' Load all documents from disk
#'
#' Loads all .automerge files from the server's data directory
#' into the documents environment.
#'
#' @param server An autosync_server object.
#'
#' @noRd
load_all_documents <- function(server) {
  files <- list.files(server$data_dir, pattern = "\\.automerge$", full.names = TRUE)
  for (f in files) {
    doc_id <- sub("\\.automerge$", "", basename(f))
    tryCatch({
      server$documents[[doc_id]] <- am_load(
        readBin(f, "raw", file.info(f)$size)
      )
    }, error = function(e) {
      warning("Failed to load document ", doc_id, ": ", conditionMessage(e))
    })
  }
}

#' Save sync state to disk
#'
#' Persists sync state for a peer with a storageId, enabling incremental
#' sync on reconnection.
#'
#' @param server Server state environment.
#' @param storage_id The peer's storage ID.
#' @param doc_id Document ID.
#' @param sync_state Automerge sync state object.
#'
#' @noRd
save_sync_state <- function(server, storage_id, doc_id, sync_state) {
  dir <- file.path(server$data_dir, ".sync_states", storage_id)
  if (!dir.exists(dir)) {
    dir.create(dir, recursive = TRUE)
  }
  path <- file.path(dir, paste0(doc_id, ".sync"))
  tryCatch(
    writeBin(am_sync_state_encode(sync_state), path),
    error = function(e) {
      warning("Failed to save sync state: ", conditionMessage(e))
    }
  )
}

#' Load persisted sync states for a peer
#'
#' Loads all persisted sync states for a peer identified by storageId.
#'
#' @param server Server state environment.
#' @param storage_id The peer's storage ID.
#' @param client_id The peer's current client ID.
#'
#' @noRd
load_sync_states <- function(server, storage_id, client_id) {
  dir <- file.path(server$data_dir, ".sync_states", storage_id)
  if (!dir.exists(dir)) {
    return(invisible())
  }
  files <- list.files(dir, pattern = "\\.sync$", full.names = TRUE)
  if (!length(files)) {
    return(invisible())
  }

  client_states <- server$sync_states[[client_id]]
  if (is.null(client_states)) {
    client_states <- new.env(hash = TRUE, parent = emptyenv())
    server$sync_states[[client_id]] <- client_states
  }

  for (f in files) {
    doc_id <- sub("\\.sync$", "", basename(f))
    tryCatch({
      sync_state <- am_sync_state_decode(
        readBin(f, "raw", file.info(f)$size)
      )
      client_states[[doc_id]] <- sync_state
    }, error = function(e) {
      warning("Failed to load sync state for ", doc_id, ": ", conditionMessage(e))
    })
  }

  invisible()
}

test_that("save_document writes file to disk", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE)
  state$data_dir <- data_dir
  state$documents <- new.env(hash = TRUE)

  doc <- automerge::am_create()
  doc_id <- "testDoc123"

  autosync:::save_document(state, doc_id, doc)

  expected_path <- file.path(data_dir, paste0(doc_id, ".automerge"))
  expect_true(file.exists(expected_path))
  expect_true(file.info(expected_path)$size > 0)
})

test_that("load_all_documents loads existing documents", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  # Create and save a document manually
  doc1 <- automerge::am_create()
  doc1_id <- "loadTest1"
  writeBin(automerge::am_save(doc1), file.path(data_dir, paste0(doc1_id, ".automerge")))

  doc2 <- automerge::am_create()
  doc2_id <- "loadTest2"
  writeBin(automerge::am_save(doc2), file.path(data_dir, paste0(doc2_id, ".automerge")))

  # Create state with empty documents env
  state <- new.env(hash = TRUE)
  state$data_dir <- data_dir
  state$documents <- new.env(hash = TRUE)

  # Load documents
  autosync:::load_all_documents(state)

  # Verify both documents were loaded
  expect_true(exists(doc1_id, envir = state$documents))
  expect_true(exists(doc2_id, envir = state$documents))
  expect_true(inherits(state$documents[[doc1_id]], "am_doc"))
  expect_true(inherits(state$documents[[doc2_id]], "am_doc"))
})

test_that("load_all_documents handles empty directory", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  state <- new.env(hash = TRUE)
  state$data_dir <- data_dir
  state$documents <- new.env(hash = TRUE)

  expect_no_error(autosync:::load_all_documents(state))
  expect_length(ls(state$documents), 0)
})

test_that("load_all_documents handles corrupted files gracefully", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  # Write invalid data
  writeBin(as.raw(c(0xFF, 0xFF, 0xFF)), file.path(data_dir, "corrupt.automerge"))

  state <- new.env(hash = TRUE)
  state$data_dir <- data_dir
  state$documents <- new.env(hash = TRUE)

  # Should warn but not error
  expect_warning(autosync:::load_all_documents(state), "Failed to load document")
  expect_length(ls(state$documents), 0)
})

test_that("save_document and load_all_documents roundtrip", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  # Create state and save a document
  state1 <- new.env(hash = TRUE)
  state1$data_dir <- data_dir
  state1$documents <- new.env(hash = TRUE)

  doc <- automerge::am_create()
  doc_id <- "roundtripTest"
  autosync:::save_document(state1, doc_id, doc)

  # Create new state and load
  state2 <- new.env(hash = TRUE)
  state2$data_dir <- data_dir
  state2$documents <- new.env(hash = TRUE)

  autosync:::load_all_documents(state2)

  expect_true(exists(doc_id, envir = state2$documents))
})

# Automerge sync client for fetching documents from remote servers

# Helper to format raw bytes as hex preview
format_hex <- function(bytes, max_len = 20L) {
  paste(
    sprintf("%02x", as.integer(bytes[1:min(max_len, length(bytes))])),
    collapse = " "
  )
}

send_msg <- function(s, msg) {
  send(s, cborenc(msg), mode = "raw", block = TRUE)
}

recv_msg <- function(s, timeout = NULL) {
  aio <- recv_aio(s, mode = "raw", timeout = timeout)
  while (unresolved(aio)) {
    run_now(1L)
  }
  aio$data
}

sync_msg <- function(type, peer_id, target_id, doc_id, data) {
  list(
    type = type,
    senderId = peer_id,
    targetId = target_id,
    documentId = doc_id,
    data = data
  )
}

apply_sync_and_reply <- function(
  s,
  doc,
  sync_state,
  data,
  peer_id,
  target_id,
  doc_id
) {
  if (length(data)) {
    tryCatch(
      am_sync_decode(doc, sync_state, data),
      error = function(e) {
        warning("am_sync_decode error: ", conditionMessage(e))
      }
    )
  }
  reply <- am_sync_encode(doc, sync_state)
  if (!is.null(reply)) {
    send_msg(s, sync_msg("sync", peer_id, target_id, doc_id, reply))
  }
  invisible()
}

bearer_headers <- function(token) {
  if (!is.null(token)) c(Authorization = paste("Bearer", token)) else NULL
}

join_msg <- function(peer_id) {
  list(
    type = "join",
    senderId = peer_id,
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )
}

#' Open a persistent sync connection
#'
#' Connects to an automerge-repo sync server and maintains a persistent
#' WebSocket connection. The connection performs the protocol handshake but
#' holds no documents on its own; open one or more live documents over it with
#' the `$open_doc()` method. Each document stays synced — receiving real-time
#' updates from other peers and flushing local changes — for as long as the
#' connection is open. Unlike [sync_fetch()], which performs a one-off
#' retrieval over a throwaway connection, several documents can share a single
#' connection here.
#'
#' @inheritParams sync_fetch
#' @param interval Interval in milliseconds for pushing local changes to
#'   the server. Default 1000. Uses [later::later()] to periodically check
#'   for and send local changes for every open document. This is a cheap no-op
#'   when there are no changes.
#'
#' @return An environment of class `"autosync_client"` with reference semantics,
#'   representing the connection:
#'   \describe{
#'     \item{`open_doc(doc_id, timeout)`}{Open a live document over this
#'       connection and return a `autosync_doc` handle for it (see below).
#'       Repeated calls for the same `doc_id` reuse the document already open
#'       on the connection rather than requesting it again.}
#'     \item{`close()`}{Disconnect and stop syncing all open documents.}
#'     \item{`active`}{Logical, whether the connection is active.}
#'   }
#'
#'   A `autosync_doc` handle returned by `$open_doc()` is itself an environment
#'   with:
#'   \describe{
#'     \item{`doc`}{The live automerge document, kept in sync with the server.}
#'     \item{`push()`}{Push this document's local changes to the server
#'       immediately.}
#'     \item{`close()`}{Stop syncing this one document (detach it from the
#'       connection); the connection and its other documents are unaffected.}
#'     \item{`active`}{Logical, whether the document is still open on an active
#'       connection.}
#'   }
#'
#' @details
#' Opening the connection performs a synchronous handshake before returning.
#' `$open_doc()` then performs a synchronous initial sync, so the returned
#' handle's `$doc` has meaningful content immediately. After that, incoming
#' changes are received asynchronously via a self-chaining promise loop, and
#' local changes are flushed periodically via a [later::later()] timer.
#'
#' Neither `close()` flushes pending local changes. Call `$push()` first if you
#' have unsynced edits — otherwise any changes made since the last
#' `sync`-interval tick may be lost.
#'
#' @examplesIf interactive()
#' server <- sync_server()
#' server$start()
#' doc_id <- create_document(server)
#'
#' conn <- sync_client(server$url)
#' doc <- conn$open_doc(doc_id)
#' automerge::am_keys(doc$doc)
#'
#' # Make local changes and push
#' automerge::am_put(doc$doc, automerge::AM_ROOT, "key", "value")
#' doc$push()
#'
#' # Open another document over the same connection
#' other <- conn$open_doc(create_document(server))
#'
#' # Disconnect (closes every document on the connection)
#' conn$close()
#' server$close()
#'
#' @export
sync_client <- function(
  url,
  timeout = 5000L,
  tls = NULL,
  token = NULL,
  interval = 1000L
) {
  peer_id <- generate_peer_id()

  s <- stream(
    dial = url,
    tls = tls,
    textframes = FALSE,
    headers = bearer_headers(token)
  )
  on.exit(close(s))

  # --- Synchronous handshake ---

  send_msg(s, join_msg(peer_id))

  peer_raw <- recv_msg(s, timeout = timeout)

  if (inherits(peer_raw, "errorValue")) {
    stop("Failed to receive peer response: ", peer_raw)
  }

  peer_msg <- cbordec(peer_raw)
  if (peer_msg$type == "error") {
    stop("Server error: ", peer_msg$message)
  }
  if (peer_msg$type != "peer") {
    stop("Expected peer message, got: ", peer_msg$type)
  }

  server_peer_id <- peer_msg$senderId

  # --- Connection state ---

  client <- new.env(parent = emptyenv())
  client$active <- TRUE
  client$stream <- s
  client$url <- url
  client$peer_id <- peer_id
  client$server_peer_id <- server_peer_id
  client$interval <- interval
  client$timeout <- timeout
  client$sync_timer <- NULL
  client$last_error <- NULL
  # Per-document registry: doc_id -> environment holding the live document, its
  # sync state, and open-sync bookkeeping. Documents are added by open_doc().
  client$documents <- new.env(parent = emptyenv())

  # Register a document on the connection, returning its registry entry.
  register_doc <- function(doc_id, doc, sync_state) {
    entry <- new.env(parent = emptyenv())
    entry$doc_id <- doc_id
    entry$doc <- doc
    entry$sync_state <- sync_state
    entry$received <- 0L
    entry$last_activity <- Sys.time()
    entry$unavailable <- FALSE
    client$documents[[doc_id]] <- entry
    entry
  }

  # Send a sync message for one document's local changes.
  send_sync_for <- function(entry) {
    sync_data <- am_sync_encode(entry$doc, entry$sync_state)
    if (!is.null(sync_data)) {
      send_msg(
        client$stream,
        sync_msg(
          "sync",
          client$peer_id,
          client$server_peer_id,
          entry$doc_id,
          sync_data
        )
      )
    }
  }

  # Process an incoming message, routing sync/doc-unavailable to the matching
  # document by its id. Messages for unknown documents are ignored.
  process_message <- function(raw_data) {
    msg <- tryCatch(cbordec(raw_data), error = function(e) NULL)
    if (is.null(msg)) {
      return()
    }

    if (msg$type == "sync" && !is.null(msg$documentId)) {
      entry <- client$documents[[msg$documentId]]
      if (!is.null(entry)) {
        entry$received <- entry$received + 1L
        entry$last_activity <- Sys.time()
        apply_sync_and_reply(
          client$stream,
          entry$doc,
          entry$sync_state,
          msg$data,
          client$peer_id,
          client$server_peer_id,
          entry$doc_id
        )
      }
    } else if (msg$type == "doc-unavailable") {
      if (!is.null(msg$documentId)) {
        entry <- client$documents[[msg$documentId]]
        if (!is.null(entry)) {
          entry$unavailable <- TRUE
        }
      }
      warning("Document not available on server: ", msg$documentId %||% "unknown")
    } else if (msg$type == "error") {
      client$last_error <- msg$message
      warning("Server error: ", msg$message)
    }
  }

  # --- Async receive loop (self-chaining promises) ---

  recv_loop <- function() {
    if (!client$active) {
      return()
    }
    aio <- recv_aio(client$stream, mode = "raw")
    # Track the pending aio so close() can settle it before tearing down the
    # stream, leaving no stale promise continuation for a later run_now to hit.
    client$recv_aio <- aio
    then(
      aio,
      onFulfilled = function(value) {
        tryCatch(
          process_message(value),
          error = function(e) {
            warning("[CLIENT] Receive error: ", conditionMessage(e))
          }
        )
        recv_loop()
      },
      onRejected = function(err) {
        client$active <- FALSE
      }
    )
    invisible()
  }

  # --- Periodic sync loop (outgoing local changes for every open document) ---

  sync_loop <- function() {
    if (!client$active) {
      return()
    }
    client$sync_timer <- later(
      function() {
        if (!client$active) {
          return()
        }
        tryCatch(
          for (id in ls(client$documents)) {
            send_sync_for(client$documents[[id]])
          },
          error = function(e) {
            client$active <- FALSE
            close(client$stream)
          }
        )
        sync_loop()
      },
      delay = interval / 1000
    )
  }

  # Build the public handle for a document registry entry.
  make_doc_handle <- function(entry) {
    handle <- new.env(parent = emptyenv())
    handle$doc_id <- entry$doc_id
    handle$doc <- entry$doc
    handle$sync_state <- entry$sync_state
    handle$stream <- client$stream
    handle$connection <- client
    handle$push <- function() send_sync_for(entry)
    handle$close <- function() {
      if (exists(entry$doc_id, envir = client$documents, inherits = FALSE)) {
        rm(list = entry$doc_id, envir = client$documents)
      }
      invisible()
    }
    # `active` tracks both the connection and whether this document is still
    # attached, so a handle whose document was closed reports inactive.
    makeActiveBinding(
      "active",
      function() {
        isTRUE(client$active) && !is.null(client$documents[[entry$doc_id]])
      },
      handle
    )
    class(handle) <- "autosync_doc"
    handle
  }

  # --- Open a live document over the connection ---

  open_doc <- function(doc_id, timeout = client$timeout) {
    if (!isTRUE(client$active)) {
      stop("Connection is not active; reconnect with sync_client()")
    }
    existing <- client$documents[[doc_id]]
    if (!is.null(existing)) {
      return(make_doc_handle(existing))
    }

    entry <- register_doc(doc_id, am_create(), am_sync_state())
    client$last_error <- NULL

    sync_data <- am_sync_encode(entry$doc, entry$sync_state)
    send_msg(
      client$stream,
      sync_msg(
        "request",
        client$peer_id,
        client$server_peer_id,
        doc_id,
        sync_data %||% raw(0)
      )
    )

    # Drive the shared event loop until the initial sync settles. Completion is
    # detected per-document: at least one sync received, then a short idle.
    deadline <- Sys.time() + timeout / 1000
    idle_secs <- 0.5
    repeat {
      run_now(0.05)

      if (!is.null(client$last_error)) {
        rm(list = doc_id, envir = client$documents)
        stop("Server error: ", client$last_error)
      }
      if (entry$unavailable) {
        rm(list = doc_id, envir = client$documents)
        stop("Document not available on server: ", doc_id)
      }
      if (!isTRUE(client$active)) {
        rm(list = doc_id, envir = client$documents)
        stop("Connection closed before document sync completed")
      }

      now <- Sys.time()
      idle <- as.numeric(now - entry$last_activity, units = "secs")
      if (entry$received > 0L && idle > idle_secs) {
        break
      }
      if (now > deadline) {
        if (entry$received == 0L) {
          rm(list = doc_id, envir = client$documents)
          stop("No sync response from server within ", timeout, "ms")
        }
        break
      }
    }

    make_doc_handle(entry)
  }

  # --- Methods ---

  client$open_doc <- open_doc

  client$close <- function() {
    if (!client$active) {
      return(invisible())
    }
    client$active <- FALSE
    if (!is.null(client$sync_timer)) {
      client$sync_timer()
      client$sync_timer <- NULL
    }
    # Settle the pending receive aio and flush its (now active == FALSE, so
    # non-rescheduling) continuation before closing the stream. This prevents a
    # stale aio external pointer from lingering in the event loop and being
    # tripped over by a later run_now after garbage collection.
    if (!is.null(client$recv_aio)) {
      stop_aio(client$recv_aio)
      for (i in seq_len(3L)) run_now()
      client$recv_aio <- NULL
    }
    close(client$stream)
    invisible()
  }

  class(client) <- "autosync_client"
  on.exit() # stream now owned by client$close

  # Start the async loops
  recv_loop()
  sync_loop()

  client
}

#' @export
print.autosync_client <- function(x, ...) {
  cat("Automerge Sync Connection\n")
  cat("  Server:", x$url, "\n")
  cat("  Documents:", length(x$documents), "\n")
  cat("  Active:", x$active, "\n")
  invisible(x)
}

#' @export
print.autosync_doc <- function(x, ...) {
  cat("Automerge Document\n")
  cat("  Document:", x$doc_id, "\n")
  cat("  Active:", x$active, "\n")
  invisible(x)
}

#' Fetch a document from a sync server
#'
#' Connects to an automerge-repo sync server and retrieves a document by ID.
#' This is useful for debugging sync issues or fetching documents from remote
#' servers like sync.automerge.org.
#'
#' @param url WebSocket URL of the sync server (e.g., "ws://localhost:3030/"
#'   or "wss://sync.automerge.org/"). Note: trailing slash may be required.
#' @param doc_id Document ID (base58check encoded string)
#' @param timeout Timeout in milliseconds for each receive operation. Default 5000.
#' @param tls (optional) for secure wss:// connections to servers with
#'   self-signed or custom CA certificates, a TLS configuration object
#'   created by [nanonext::tls_config()].
#' @param token (optional) JWT (ID token) for authenticated servers.
#'   Sent as a Bearer token in the Authorization header of the WebSocket
#'   upgrade request.
#' @param verbose Logical, print debug messages. Default FALSE.
#'
#' @return An automerge document object containing the fetched data.
#'
#' @details
#' The function implements the automerge-repo sync protocol:
#' 1. Connects via WebSocket
#' 2. Sends join message with peer ID
#' 3. Receives peer response from server
#' 4. Sends sync request for the specified document
#' 5. Receives and applies sync messages until complete
#'
#' Sync is considered complete when no new messages arrive within the timeout
#' after at least one sync round.
#'
#' @examplesIf interactive()
#' # Fetch from public sync server
#' doc <- sync_fetch("wss://sync.automerge.org", "4F63WJPDzbHkkfKa66h1Qrr1sC5U")
#'
#' # Fetch from local server with debug output
#' doc <- sync_fetch(server$url, "myDocId", verbose = TRUE)
#'
#' # Fetch from server with self-signed certificate
#' cert <- nanonext::write_cert()
#' tls <- nanonext::tls_config(client = cert$client)
#' doc <- sync_fetch(server$url, "myDocId", tls = tls)
#'
#' # Fetch from authenticated server
#' doc <- sync_fetch(
#'   "wss://secure.example.com",
#'   "myDocId",
#'   token = "eyJhbGciOi..."
#' )
#'
#' # Inspect the document
#' automerge::am_keys(doc)
#'
#' @export
sync_fetch <- function(
  url,
  doc_id,
  timeout = 5000L,
  tls = NULL,
  token = NULL,
  verbose = FALSE
) {
  doc <- am_create()
  sync_state <- am_sync_state()

  peer_id <- generate_peer_id()

  if (verbose) {
    message("[CLIENT] Connecting to ", url)
  }
  if (verbose) {
    message("[CLIENT] Our peer ID: ", peer_id)
  }
  if (verbose) {
    message("[CLIENT] Requesting document: ", doc_id)
  }

  s <- stream(
    dial = url,
    tls = tls,
    textframes = FALSE,
    headers = bearer_headers(token)
  )
  on.exit(close(s))

  if (verbose) {
    message("[CLIENT] Connected, sending join message")
  }

  join_bytes <- cborenc(join_msg(peer_id))
  if (verbose) {
    message(
      "[CLIENT] Sending join (",
      length(join_bytes),
      " bytes): ",
      format_hex(join_bytes)
    )
  }
  send(s, join_bytes, mode = "raw", block = TRUE)

  if (verbose) {
    message("[CLIENT] Waiting for peer response...")
  }
  peer_raw <- recv_msg(s, timeout = timeout)

  if (inherits(peer_raw, "errorValue")) {
    stop("Failed to receive peer response: ", peer_raw)
  }

  if (verbose) {
    message(
      "[CLIENT] Received ",
      length(peer_raw),
      " bytes: ",
      format_hex(peer_raw)
    )
  }

  peer_msg <- cbordec(peer_raw)

  if (verbose) {
    message("[CLIENT] Message type: ", peer_msg$type)
  }

  if (peer_msg$type == "error") {
    stop("Server error: ", peer_msg$message)
  }

  if (peer_msg$type != "peer") {
    stop("Expected peer message, got: ", peer_msg$type)
  }

  server_peer_id <- peer_msg$senderId
  if (verbose) {
    message("[CLIENT] Server peer ID: ", server_peer_id)
    message("[CLIENT] Protocol version: ", peer_msg$selectedProtocolVersion)
  }

  sync_data <- am_sync_encode(doc, sync_state)
  if (verbose) {
    message(
      "[CLIENT] Initial sync data: ",
      if (is.null(sync_data)) "NULL" else paste(length(sync_data), "bytes")
    )
  }

  request_bytes <- cborenc(
    sync_msg("request", peer_id, server_peer_id, doc_id, sync_data %||% raw(0))
  )
  if (verbose) {
    message("[CLIENT] Sending request (", length(request_bytes), " bytes)")
  }
  send(s, request_bytes, mode = "raw", block = TRUE)

  # Sync loop - receive and process messages until timeout
  sync_rounds <- 0L
  idle_timeout <- 2000L # Wait 2 seconds for additional messages after sync

  repeat {
    result <- recv_msg(s, timeout = idle_timeout)

    if (inherits(result, "errorValue")) {
      # Timeout or error - no more messages
      if (verbose) {
        message("[CLIENT] No more messages (timeout)")
      }
      break
    }

    if (verbose) {
      message(
        "[CLIENT] Received ",
        length(result),
        " bytes: ",
        format_hex(result)
      )
    }

    msg <- tryCatch(
      cbordec(result),
      error = function(e) {
        if (verbose) {
          message("[CLIENT] CBOR decode error: ", conditionMessage(e))
        }
        return(NULL)
      }
    )

    if (is.null(msg)) {
      next
    }

    if (verbose) {
      message("[CLIENT] Message type: ", msg$type)
    }

    if (
      msg$type == "sync" && !is.null(msg$documentId) && msg$documentId != doc_id
    ) {
      if (verbose) {
        message(
          "[CLIENT] Ignoring sync for different document: ",
          msg$documentId
        )
      }
      next
    }

    if (msg$type == "sync") {
      sync_rounds <- sync_rounds + 1L
      data_len <- if (is.null(msg$data)) 0L else length(msg$data)
      if (verbose) {
        message(
          "[CLIENT] Sync round ",
          sync_rounds,
          " - received ",
          data_len,
          " bytes of sync data"
        )
      }

      if (length(msg$data)) {
        if (verbose) {
          message("[CLIENT] Sync data hex: ", format_hex(msg$data, 30L))
        }
        tryCatch(
          {
            am_sync_decode(doc, sync_state, msg$data)
            if (verbose) message("[CLIENT] Applied sync data successfully")
          },
          error = function(e) {
            warning("[CLIENT] am_sync_decode error: ", conditionMessage(e))
          }
        )
      }

      if (verbose) {
        keys <- tryCatch(am_keys(doc), error = function(e) character(0))
        message(
          "[CLIENT] Document now has ",
          length(keys),
          " keys: ",
          paste(keys[1:min(5, length(keys))], collapse = ", "),
          if (length(keys) > 5) "..." else ""
        )
      }

      reply_data <- am_sync_encode(doc, sync_state)
      if (!is.null(reply_data)) {
        if (verbose) {
          message("[CLIENT] Sending ", length(reply_data), " bytes in response")
        }
        send_msg(
          s,
          sync_msg("sync", peer_id, server_peer_id, doc_id, reply_data)
        )
      } else {
        if (verbose) message("[CLIENT] No more data to send from our side")
      }
    } else if (msg$type == "error") {
      stop("Server error: ", msg$message)
    } else if (msg$type == "doc-unavailable") {
      stop("Document not available on server: ", doc_id)
    } else {
      if (verbose) message("[CLIENT] Ignoring message type: ", msg$type)
    }
  }

  if (sync_rounds == 0L) {
    warning("No sync messages received - document may be empty or unavailable")
  }

  if (verbose) {
    message("[CLIENT] Fetch complete after ", sync_rounds, " sync rounds")
    str(doc)
  }

  doc
}

test_that("format_hex formats raw bytes as hex", {
  format_hex <- autosync:::format_hex

  # Simple test
  bytes <- as.raw(c(0xDE, 0xAD, 0xBE, 0xEF))
  result <- format_hex(bytes)
  expect_equal(result, "de ad be ef")

  # Test with max_len limit
  bytes <- as.raw(1:10)
  result <- format_hex(bytes, max_len = 5L)
  expect_equal(result, "01 02 03 04 05")

  # Test with single byte
  bytes <- as.raw(0xFF)
  result <- format_hex(bytes)
  expect_equal(result, "ff")
})

# Test helpers shared between sync_fetch() and sync_client() tests

fake_peer <- secretbase::cborenc(list(
  type = "peer",
  senderId = "server123",
  selectedProtocolVersion = "1"
))

# Mock recv_aio that returns a scripted sequence of responses, one per call.
# Each element of `responses` is either an errorValue or raw CBOR bytes.
scripted_recv <- function(responses) {
  i <- 0L
  function(...) {
    i <<- i + 1L
    list(data = responses[[min(i, length(responses))]])
  }
}

# sync_fetch() tests

test_that("sync_fetch retrieves document from server", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  # Create and populate a document

  doc_id <- create_document(server)
  doc <- get_document(server, doc_id)
  automerge::am_put(doc, automerge::AM_ROOT, "test_key", "test_value")
  automerge::am_put(doc, automerge::AM_ROOT, "number", 42L)
  automerge::am_put(doc, automerge::AM_ROOT, "flag", TRUE)

  # Fetch the document
  fetched <- sync_fetch(server$url, doc_id, timeout = 5000L, verbose = FALSE)

  expect_true(inherits(fetched, "am_doc"))
  expect_equal(automerge::am_get(fetched, automerge::AM_ROOT, "test_key"), "test_value")
  expect_equal(automerge::am_get(fetched, automerge::AM_ROOT, "number"), 42L)
  expect_equal(automerge::am_get(fetched, automerge::AM_ROOT, "flag"), TRUE)
})

test_that("sync_fetch works in verbose mode", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  # Create and populate a document
  doc_id <- create_document(server)
  doc <- get_document(server, doc_id)
  automerge::am_put(doc, automerge::AM_ROOT, "verbose_test", "value")

  # Fetch with verbose = TRUE and capture output
  output <- capture.output({
    fetched <- sync_fetch(server$url, doc_id, timeout = 5000L, verbose = TRUE)
  }, type = "message")

  # Should have verbose output
  expect_true(any(grepl("\\[CLIENT\\]", output)))
  expect_true(any(grepl("Connecting", output)))
  expect_true(any(grepl("peer", output, ignore.case = TRUE)))
  expect_true(any(grepl("join", output, ignore.case = TRUE)))
  expect_true(any(grepl("sync", output, ignore.case = TRUE)))
  expect_true(any(grepl("document", output, ignore.case = TRUE)))

  # Document should still be fetched correctly
  expect_true(inherits(fetched, "am_doc"))
  expect_equal(automerge::am_get(fetched, automerge::AM_ROOT, "verbose_test"), "value")
})

test_that("sync_fetch non-verbose mode produces no output", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)

  output <- capture.output({
    fetched <- sync_fetch(server$url, doc_id, timeout = 5000L, verbose = FALSE)
  }, type = "message")

  # Should have no verbose output

  expect_length(output, 0)
})

test_that("sync_fetch sends access_token as Authorization header", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)
  doc <- get_document(server, doc_id)
  automerge::am_put(doc, automerge::AM_ROOT, "key", "value")

  # Server has no auth, so token is sent but ignored
  fetched <- sync_fetch(server$url, doc_id, token = "test_token_12345_extra_padding")

  expect_true(inherits(fetched, "am_doc"))
  expect_equal(automerge::am_get(fetched, automerge::AM_ROOT, "key"), "value")
})

test_that("sync_fetch errors when peer response fails", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(structure(5L, class = "errorValue"))),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_fetch("ws://fake:1234", "fake_doc_id"),
    "Failed to receive peer response"
  )
})

test_that("sync_fetch errors when server returns error during handshake", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      secretbase::cborenc(list(type = "error", message = "auth failed"))
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_fetch("ws://fake:1234", "fake_doc_id"),
    "Server error: auth failed"
  )
})

test_that("sync_fetch errors on unexpected message type", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      secretbase::cborenc(list(type = "sync", senderId = "server123"))
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_fetch("ws://fake:1234", "fake_doc_id"),
    "Expected peer message, got: sync"
  )
})

test_that("sync_fetch errors when server returns error during sync", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      fake_peer,
      secretbase::cborenc(list(type = "error", message = "sync failed"))
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_fetch("ws://fake:1234", "fake_doc_id"),
    "Server error: sync failed"
  )
})

test_that("sync_fetch errors when document is unavailable", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      fake_peer,
      secretbase::cborenc(list(type = "doc-unavailable"))
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_fetch("ws://fake:1234", "fake_doc_id"),
    "Document not available on server: fake_doc_id"
  )
})

test_that("sync_fetch warns when no sync messages received", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      fake_peer,
      structure(5L, class = "errorValue")
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_warning(
    sync_fetch("ws://fake:1234", "fake_doc_id"),
    "No sync messages received"
  )
})

test_that("sync_fetch verbose mode logs CBOR decode errors", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      fake_peer,
      as.raw(c(0xFF, 0xFF, 0xFF)),
      structure(5L, class = "errorValue")
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  output <- capture.output(
    suppressWarnings(
      sync_fetch("ws://fake:1234", "fake_doc_id", verbose = TRUE)
    ),
    type = "message"
  )
  expect_true(any(grepl("CBOR decode error", output)))
})

test_that("sync_fetch verbose mode reports skipped foreign sync messages", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      fake_peer,
      secretbase::cborenc(list(
        type = "sync",
        documentId = "different_doc",
        data = raw(0)
      )),
      structure(5L, class = "errorValue")
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  output <- capture.output(
    suppressWarnings(
      sync_fetch("ws://fake:1234", "fake_doc_id", verbose = TRUE)
    ),
    type = "message"
  )
  expect_true(any(grepl("Ignoring sync for different document", output)))
})

test_that("sync_fetch verbose mode reports unknown message types", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      fake_peer,
      secretbase::cborenc(list(type = "ephemeral", data = raw(0))),
      structure(5L, class = "errorValue")
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  output <- capture.output(
    suppressWarnings(
      sync_fetch("ws://fake:1234", "fake_doc_id", verbose = TRUE)
    ),
    type = "message"
  )
  expect_true(any(grepl("Ignoring message type: ephemeral", output)))
})

# sync_client() tests

test_that("sync_client errors when peer response fails", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(structure(5L, class = "errorValue"))),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_client("ws://fake:1234"),
    "Failed to receive peer response"
  )
})

test_that("sync_client errors when server returns error during handshake", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      secretbase::cborenc(list(type = "error", message = "auth failed"))
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_client("ws://fake:1234"),
    "Server error: auth failed"
  )
})

test_that("sync_client errors on unexpected message type during handshake", {
  local_mocked_bindings(
    stream = function(...) rawConnection(raw(0)),
    send = function(...) invisible(NULL),
    recv_aio = scripted_recv(list(
      secretbase::cborenc(list(type = "sync", senderId = "server123"))
    )),
    unresolved = function(...) FALSE,
    run_now = function(...) invisible(NULL)
  )

  expect_error(
    sync_client("ws://fake:1234"),
    "Expected peer message, got: sync"
  )
})

test_that("open_doc errors when the server reports an error", {
  skip_on_cran()
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url)
  on.exit(if (conn$active) conn$close(), add = TRUE)

  # An unparseable document ID makes the server send an error and disconnect.
  expect_error(
    suppressWarnings(conn$open_doc("not a valid id!")),
    "Server error"
  )
})

test_that("open_doc errors when the document is unavailable", {
  skip_on_cran()
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  # share = FALSE denies all document requests (sends doc-unavailable).
  server <- sync_server(data_dir = data_dir, share = FALSE)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url)
  on.exit(if (conn$active) conn$close(), add = TRUE)

  expect_error(
    suppressWarnings(conn$open_doc(generate_document_id())),
    "Document not available on server"
  )
})

test_that("open_doc times out without a sync response", {
  skip_on_cran()
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url)
  on.exit(if (conn$active) conn$close(), add = TRUE)

  # Drop the outgoing request so the server never replies.
  local_mocked_bindings(send_msg = function(s, msg) invisible(NULL))

  expect_error(
    conn$open_doc(generate_document_id(), timeout = 200L),
    "No sync response from server"
  )
})

test_that("sync_client connects and open_doc receives document", {
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)
  doc <- get_document(server, doc_id)
  automerge::am_put(doc, automerge::AM_ROOT, "greeting", "hello")

  conn <- sync_client(server$url)
  on.exit(conn$close(), add = TRUE)

  expect_s3_class(conn, "autosync_client")
  expect_true(conn$active)

  handle <- conn$open_doc(doc_id)
  expect_s3_class(handle, "autosync_doc")
  expect_true(handle$active)
  expect_equal(handle$doc_id, doc_id)
  expect_equal(
    automerge::am_get(handle$doc, automerge::AM_ROOT, "greeting"),
    "hello"
  )
})

test_that("a document handle's $close() detaches it from the connection", {
  skip_on_cran()
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)
  conn <- sync_client(server$url)
  on.exit(conn$close(), add = TRUE)

  handle <- conn$open_doc(doc_id)
  expect_true(handle$active)

  # Closing the handle drops the document so the handle reports inactive.
  handle$close()
  expect_false(handle$active)

  # A second close is an idempotent no-op.
  expect_no_error(handle$close())
})

test_that("open_doc errors once the connection is closed", {
  skip_on_cran()
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url)
  conn$close()

  expect_error(
    conn$open_doc(generate_document_id()),
    "not active"
  )
})

test_that("open_doc reuses one connection across documents", {
  skip_on_cran()
  drain_later()
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  id1 <- create_document(server)
  id2 <- create_document(server)

  conn <- sync_client(server$url)
  on.exit(conn$close(), add = TRUE)

  conns_before <- length(ls(attr(server, "sync")$connections))
  h1 <- conn$open_doc(id1)
  h2 <- conn$open_doc(id2)
  for (i in seq_len(10)) later::run_now(0.05)

  # Opening documents reuses the same WebSocket, not a new server connection.
  expect_equal(length(ls(attr(server, "sync")$connections)), conns_before)
  expect_identical(h1$stream, conn$stream)
  expect_identical(h2$stream, conn$stream)
  expect_identical(h1$connection, conn)

  # Re-opening the same document returns a handle backed by the same live doc.
  expect_identical(conn$open_doc(id1)$doc, h1$doc)
})

test_that("open_doc handle $push() pushes local changes to server", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)

  conn <- sync_client(server$url)
  on.exit(conn$close(), add = TRUE)
  handle <- conn$open_doc(doc_id)

  # Make a local change and push
  automerge::am_put(handle$doc, automerge::AM_ROOT, "from_client", "value1")
  handle$push()

  # Give server time to process
  for (i in seq_len(20)) later::run_now(0.1)

  server_doc <- get_document(server, doc_id)
  expect_equal(
    automerge::am_get(server_doc, automerge::AM_ROOT, "from_client"),
    "value1"
  )
})

test_that("open_doc handle receives server-side changes", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)

  conn <- sync_client(server$url)
  on.exit(conn$close(), add = TRUE)
  handle <- conn$open_doc(doc_id)

  # Make a change on the server side and sync to client via a second client
  server_doc <- get_document(server, doc_id)
  automerge::am_put(server_doc, automerge::AM_ROOT, "server_key", "server_val")

  # Use a second fetch to trigger broadcast_sync on the server
  fetched <- sync_fetch(server$url, doc_id, timeout = 2000L)

  # Give the async receive loop time to process
  for (i in seq_len(30)) later::run_now(0.1)

  expect_equal(
    automerge::am_get(handle$doc, automerge::AM_ROOT, "server_key"),
    "server_val"
  )
})

test_that("sync_client deactivates when periodic sync errors", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- generate_document_id()

  conn <- sync_client(server$url, interval = 50L)
  on.exit(if (conn$active) conn$close(), add = TRUE)
  handle <- conn$open_doc(doc_id)

  # Give the periodic sync something to send, then force send to fail
  # to trigger the timer's error handler.
  automerge::am_put(handle$doc, automerge::AM_ROOT, "k", "v")
  local_mocked_bindings(send_msg = function(s, msg) stop("simulated"))

  while (!later::loop_empty()) later::run_now(1L)

  expect_false(conn$active)
})

test_that("sync_client async loop survives process_message errors", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- generate_document_id()

  conn <- sync_client(server$url, interval = 999000L)
  on.exit(conn$close(), add = TRUE)
  handle <- conn$open_doc(doc_id)

  # Force every subsequent apply_sync_and_reply to throw so the async
  # receive loop has to recover.
  local_mocked_bindings(
    apply_sync_and_reply = function(...) stop("synthetic")
  )

  # Trigger a server-side broadcast to push a sync at our client
  server_doc <- get_document(server, doc_id)
  automerge::am_put(server_doc, automerge::AM_ROOT, "trigger", "v")
  expect_warning(
    sync_fetch(server$url, doc_id, timeout = 2000L),
    "Receive error: synthetic"
  )

  # Wait for the recv to fire, snapshot, then close so the loop can drain
  # (sync_loop reschedules itself, so loop_empty() never goes true while active).
  later::run_now(2L)
  active_after_error <- conn$active
  conn$close()
  while (!later::loop_empty()) later::run_now(1L)

  expect_true(active_after_error)
})

test_that("apply_sync_and_reply warns on decode error", {
  doc <- automerge::am_create()
  sync_state <- automerge::am_sync_state()

  local_mocked_bindings(send_msg = function(s, msg) invisible(NULL))

  expect_warning(
    apply_sync_and_reply(
      s = NULL,
      doc = doc,
      sync_state = sync_state,
      data = as.raw(c(0xFF, 0xFF, 0xFF)),
      peer_id = "p",
      target_id = "t",
      doc_id = "d"
    ),
    "am_sync_decode error"
  )
})

test_that("process_message handles non-sync server messages", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url, interval = 999000L)
  on.exit(conn$close(), add = TRUE)

  process_message <- environment(conn$open_doc)$process_message

  expect_warning(
    process_message(secretbase::cborenc(list(type = "error", message = "boom"))),
    "Server error: boom"
  )
  expect_warning(
    process_message(secretbase::cborenc(list(type = "doc-unavailable"))),
    "Document not available"
  )
  # Malformed CBOR is swallowed (process_message returns silently)
  expect_silent(process_message(as.raw(c(0xFF, 0xFF))))
})

test_that("sync_client periodic timer is a no-op when connection becomes inactive", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url, interval = 100L)
  on.exit(if (conn$active) conn$close(), add = TRUE)

  # Close the stream externally; the pending recv promise will reject,
  # flipping active to FALSE without cancelling the still-scheduled timer.
  # The next timer tick must hit the !active early return.
  close(conn$stream)

  deadline <- Sys.time() + 1
  while (Sys.time() < deadline) later::run_now(0.05)

  expect_false(conn$active)
})

test_that("sync_client $close() stops the connection", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  conn <- sync_client(server$url)
  expect_true(conn$active)

  conn$close()
  expect_false(conn$active)

  # Calling close again is a no-op
  expect_silent(conn$close())
})

test_that("print methods display connection and document info", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()
  on.exit(server$close(), add = TRUE)

  doc_id <- create_document(server)

  conn <- sync_client(server$url)
  on.exit(conn$close(), add = TRUE)

  conn_out <- capture.output(print(conn))
  expect_true(any(grepl("Automerge Sync Connection", conn_out)))
  expect_true(any(grepl("Active: TRUE", conn_out)))

  doc_out <- capture.output(print(conn$open_doc(doc_id)))
  expect_true(any(grepl("Automerge Document", doc_out)))
  expect_true(any(grepl(doc_id, doc_out, fixed = TRUE)))
  expect_true(any(grepl("Active: TRUE", doc_out)))
})

test_that("sync_fetch returns empty document for new document ID", {
  data_dir <- tempfile()
  dir.create(data_dir)
  on.exit(unlink(data_dir, recursive = TRUE))

  server <- sync_server(data_dir = data_dir)
  server$start()

  on.exit(server$close(), add = TRUE)

  # Request a new document ID (server creates empty doc)
  new_id <- generate_document_id()

  fetched <- sync_fetch(server$url, new_id, timeout = 2000L, verbose = FALSE)

  # Should return a valid but empty document
  expect_true(inherits(fetched, "am_doc"))
  expect_length(automerge::am_keys(fetched, automerge::AM_ROOT), 0)
})

# Utility functions for autosync

#' Generate a new document ID
#'
#' Creates a new unique document ID compatible with automerge-repo.
#' The ID is a 16-byte random value encoded with Base58Check.
#'
#' @return Character string (Base58Check encoded).
#'
#' @examples
#' generate_document_id()
#'
#' @export
generate_document_id <- function() base58enc(random(16L, convert = FALSE))

#' Generate a peer ID
#'
#' Creates a new unique peer ID compatible with automerge-repo.
#' The ID is 16 random bytes, Base64-encoded.
#'
#' @return Character string (Base64 encoded).
#'
#' @noRd
generate_peer_id <- function() base64enc(random(16L, convert = FALSE))

#' Close a WebSocket connection
#'
#' Closes a WebSocket connection by ID.
#'
#' @param server Server state environment.
#' @param id Connection ID (temp_id or client_id).
#'
#' @noRd
close_connection <- function(server, id) {
  conn <- server$connections[[id]]
  if (!is.null(conn) && !is.null(conn$ws)) {
    conn$ws$close()
  }
}

test_that("generate_peer_id creates valid Base64 encoded IDs", {
  id <- autosync:::generate_peer_id()
  expect_type(id, "character")
  expect_true(nchar(id) > 0)
  # Should be decodable as Base64 to 16 bytes
  bytes <- secretbase::base64dec(id, convert = FALSE)
  expect_length(bytes, 16L)
})

test_that("generate_peer_id creates unique IDs", {
  ids <- replicate(100, autosync:::generate_peer_id())
  expect_equal(length(unique(ids)), 100)
})

test_that("generate_document_id creates unique IDs", {
  ids <- replicate(100, generate_document_id())
  expect_equal(length(unique(ids)), 100)
})

# ---- close_connection tests ----

test_that("close_connection closes ws for existing connection", {
  server <- new.env(hash = TRUE, parent = emptyenv())
  server$connections <- new.env(hash = TRUE, parent = emptyenv())

  ws <- new.env(hash = TRUE)
  ws$closed <- FALSE
  ws$close <- function() ws$closed <- TRUE

  server$connections[["conn1"]] <- list(ws = ws, client_id = "conn1")

  autosync:::close_connection(server, "conn1")
  expect_true(ws$closed)
})

test_that("close_connection handles missing connection gracefully", {
  server <- new.env(hash = TRUE, parent = emptyenv())
  server$connections <- new.env(hash = TRUE, parent = emptyenv())

  expect_no_error(autosync:::close_connection(server, "nonexistent"))
})

test_that("close_connection handles connection with NULL ws", {
  server <- new.env(hash = TRUE, parent = emptyenv())
  server$connections <- new.env(hash = TRUE, parent = emptyenv())

  server$connections[["conn2"]] <- list(ws = NULL, client_id = "conn2")

  expect_no_error(autosync:::close_connection(server, "conn2"))
})

#' autosync: 'Automerge' Sync Server and Client
#'
#' A WebSocket-based implementation of the 'automerge-repo' synchronization
#' protocol used by 'sync.automerge.org'. Acts as a sync server, enabling 'R' to
#' serve as a synchronization hub for 'Automerge' clients in 'JavaScript',
#' 'Rust', and other languages, and as a client for fetching, editing, and
#' synchronizing documents hosted on remote servers.
#'
#' @section Main Functions:
#' \describe{
#'   \item{[sync_server()]}{Create a new sync server with `$start()` and
#'     `$close()` methods}
#' }
#'
#' @section Document Management:
#' \describe{
#'   \item{[create_document()]}{Create a new document}
#'   \item{[get_document()]}{Retrieve a document by ID}
#'   \item{[list_documents()]}{List all document IDs}
#'   \item{[generate_document_id()]}{Generate a new document ID}
#' }
#'
#' @section Protocol:
#' The server implements the automerge-repo sync protocol over WebSockets.
#' Messages are CBOR-encoded and include:
#' \describe{
#'   \item{join/peer}{Handshake messages for connection establishment}
#'   \item{request/sync}{Document synchronization messages}
#'   \item{ephemeral}{Transient messages forwarded without persistence}
#'   \item{error}{Error notifications}
#' }
#'
#' @section Example:
#' \preformatted{
#' # Create and start a server
#' server <- sync_server()
#' server$start()
#' server$url
#'
#' # Stop when done
#' server$close()
#' }
#'
#' @docType package
#' @name autosync-package
#' @aliases autosync
#'
#' @importFrom automerge am_create am_get am_keys am_length am_load am_save
#'   am_sync_decode am_sync_encode am_sync_state am_sync_state_decode
#'   am_sync_state_encode AM_ROOT
#' @importFrom httr2 oauth_client oauth_flow_auth_code oauth_redirect_uri
#'   oauth_server_metadata
#' @importFrom jose jwt_decode_sig read_jwk
#' @importFrom later later run_now
#' @importFrom nanonext handler_ws http_server is_error_value ncurl random recv
#'   recv_aio send stop_aio stream tls_config unresolved write_cert
#' @importFrom promises then
#' @importFrom secretbase base64enc base64dec base58enc base58dec cborenc
#'   cbordec jsondec jsonenc
#' @importFrom utils str
"_PACKAGE"

# Message handlers for autosync

#' Evaluate the share policy for a client and document
#'
#' @param share The share policy (TRUE, FALSE, NA, or function).
#' @param client_id Client's peer ID.
#' @param doc_id Document ID string.
#'
#' @return TRUE (announce and allow), NA (allow only), or FALSE (deny).
#'
#' @noRd
eval_share <- function(share, client_id, doc_id) {
  if (is.function(share)) share(client_id, doc_id) else share
}

#' Main message dispatcher
#'
#' Routes incoming CBOR messages to the appropriate handler.
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID (or temp_id if pre-handshake).
#' @param temp_id Temporary connection ID.
#' @param raw_msg Raw CBOR message bytes.
#'
#' @noRd
handle_message <- function(server, client_id, temp_id, raw_msg) {
  if (!length(raw_msg)) {
    return(invisible())
  }

  msg <- tryCatch(
    cbordec(raw_msg),
    error = function(e) {
      warning("CBOR decode error from ", client_id, ": ", conditionMessage(e))
      return(NULL)
    }
  )
  if (is.null(msg)) {
    return(invisible())
  }

  conn <- server$connections[[temp_id]]
  if (is.null(conn$client_id) && msg$type != "join") {
    send_error(server, msg$senderId %||% "unknown", "Expected join message", temp_id = temp_id)
    close_connection(server, temp_id)
    return(invisible())
  }

  switch(
    msg$type,
    "join" = handle_join(server, temp_id, msg),
    "request" = handle_sync(server, client_id, msg, is_request = TRUE),
    "sync" = handle_sync(server, client_id, msg, is_request = FALSE),
    "leave" = handle_leave(server, client_id, msg),
    "ephemeral" = handle_ephemeral(server, client_id, msg),
    "error" = handle_error(server, client_id, msg),
    warning("Unknown message type: ", msg$type)
  )

  invisible()
}

#' Handle join handshake message
#'
#' @param server An autosync_server object.
#' @param temp_id Temporary connection ID.
#' @param msg Decoded join message.
#'
#' @noRd
handle_join <- function(server, temp_id, msg) {
  # Validate protocol version
  if (
    !is.null(msg$supportedProtocolVersions) &&
      !("1" %in% msg$supportedProtocolVersions)
  ) {
    send_error(
      server,
      msg$senderId,
      "Unsupported protocol version",
      temp_id = temp_id
    )
    close_connection(server, temp_id)
    return(invisible())
  }

  # Close previous connection from same client (reconnection)
  client_id <- msg$senderId
  if (exists(client_id, envir = server$connections, inherits = FALSE)) {
    old_conn <- server$connections[[client_id]]
    if (
      !is.null(old_conn$ws) &&
        !identical(old_conn$ws, server$connections[[temp_id]]$ws)
    ) {
      old_temp_id <- as.character(old_conn$ws$id)
      handle_disconnect(server, client_id)
      if (exists(old_temp_id, envir = server$connections, inherits = FALSE)) {
        rm(list = old_temp_id, envir = server$connections)
      }
      if (exists(client_id, envir = server$connections, inherits = FALSE)) {
        rm(list = client_id, envir = server$connections)
      }
      old_conn$ws$close()
    }
  }

  # Update connection: store client_id and metadata
  # Must update the original entry (R lists are copied, not referenced)
  server$connections[[temp_id]]$client_id <- client_id
  server$connections[[temp_id]]$metadata <- msg$peerMetadata %||% list()

  # Also store connection indexed by client_id for message routing
  server$connections[[client_id]] <- server$connections[[temp_id]]

  # Load persisted sync states for peers with a storageId
  peer_storage_id <- msg$peerMetadata$storageId
  if (!is.null(peer_storage_id) && is.character(peer_storage_id)) {
    load_sync_states(server, peer_storage_id, client_id)
  }

  response <- list(
    type = "peer",
    senderId = server$peer_id,
    targetId = client_id,
    peerMetadata = if (is.null(server$storage_id)) {
      list(isEphemeral = TRUE)
    } else {
      list(storageId = server$storage_id, isEphemeral = FALSE)
    },
    selectedProtocolVersion = "1"
  )
  send_to_peer(server, client_id, response)

  for (doc_id in ls(server$documents)) {
    if (isTRUE(eval_share(server$share, client_id, doc_id))) {
      sync_to_peer(server, client_id, doc_id, server$documents[[doc_id]])
    }
  }

  invisible()
}

#' Sync a single document to a single peer
#'
#' Creates sync state if needed and sends a sync message.
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#' @param doc_id Document ID.
#' @param doc Automerge document object.
#'
#' @noRd
sync_to_peer <- function(server, client_id, doc_id, doc) {
  client_states <- server$sync_states[[client_id]]
  if (is.null(client_states)) {
    client_states <- new.env(hash = TRUE, parent = emptyenv())
    server$sync_states[[client_id]] <- client_states
  }
  if (is.null(client_states[[doc_id]])) {
    client_states[[doc_id]] <- am_sync_state()
    add_doc_peer(server, doc_id, client_id)
  }

  sync_data <- tryCatch(
    am_sync_encode(doc, client_states[[doc_id]]),
    error = function(e) NULL
  )
  if (!is.null(sync_data)) {
    send_to_peer(server, client_id, list(
      type = "sync",
      senderId = server$peer_id,
      targetId = client_id,
      documentId = doc_id,
      data = sync_data
    ))
  }
  invisible()
}

#' Handle sync or request message
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#' @param msg Decoded sync/request message.
#' @param is_request Logical, TRUE for request messages.
#'
#' @noRd
handle_sync <- function(server, client_id, msg, is_request) {
  # Validate targetId matches this server
  if (!is.null(msg$targetId) && msg$targetId != server$peer_id) {
    return(invisible())
  }

  doc_id <- msg$documentId

  # Validate document ID format (base58check)
  doc_bytes <- tryCatch(
    base58dec(doc_id, convert = FALSE),
    error = function(e) NULL
  )
  if (is.null(doc_bytes)) {
    send_error(server, client_id, "Invalid document ID format")
    close_connection(server, client_id)
    return(invisible())
  }

  # Check share policy for access control
  if (isFALSE(eval_share(server$share, client_id, doc_id))) {
    send_unavailable(server, client_id, doc_id)
    return(invisible())
  }

  # Get or create document based on server policy
  doc <- server$documents[[doc_id]]

  if (is.null(doc)) {
    if (server$auto_create_docs) {
      doc <- am_create()
      server$documents[[doc_id]] <- doc
    } else {
      send_unavailable(server, client_id, doc_id)
      return(invisible())
    }
  }

  # Get or create sync state for this client/document pair
  # Uses nested envs: sync_states[[client_id]][[doc_id]]
  client_states <- server$sync_states[[client_id]]
  if (is.null(client_states)) {
    client_states <- new.env(hash = TRUE, parent = emptyenv())
    server$sync_states[[client_id]] <- client_states
  }
  sync_state <- client_states[[doc_id]]
  if (is.null(sync_state)) {
    sync_state <- am_sync_state()
    client_states[[doc_id]] <- sync_state
    add_doc_peer(server, doc_id, client_id)
  }

  if (length(msg$data)) {
    tryCatch(
      am_sync_decode(doc, sync_state, msg$data),
      error = function(e) {
        warning("am_sync_decode error: ", conditionMessage(e))
      }
    )
    save_document(server, doc_id, doc)

    # Persist sync state for peers with a storageId
    conn <- server$connections[[client_id]]
    peer_storage_id <- conn$metadata$storageId
    if (!is.null(peer_storage_id) && is.character(peer_storage_id)) {
      save_sync_state(server, peer_storage_id, doc_id, sync_state)
    }
  }

  reply_data <- tryCatch(
    am_sync_encode(doc, sync_state),
    error = function(e) NULL
  )
  if (!is.null(reply_data)) {
    response <- list(
      type = "sync",
      senderId = server$peer_id,
      targetId = client_id,
      documentId = doc_id,
      data = reply_data
    )
    send_to_peer(server, client_id, response)
  }

  broadcast_sync(server, client_id, doc_id, doc)

  invisible()
}

#' Handle leave message
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#' @param msg Decoded leave message.
#'
#' @noRd
handle_leave <- function(server, client_id, msg) {
  # Client is disconnecting gracefully
  # Clean up will happen in onClose handler
  invisible()
}

#' Handle ephemeral message
#'
#' Forwards ephemeral messages without persistence. Supports both:
#' - Point-to-point: message with targetId forwards to specific peer
#' - Broadcast: message with documentId relays to all peers on that document
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#' @param msg Decoded ephemeral message.
#'
#' @noRd
handle_ephemeral <- function(server, client_id, msg) {
  session_id <- msg$sessionId
  count <- msg$count
  if (!is.null(session_id) && !is.null(count)) {
    key <- paste0(client_id, "\x1f", session_id)
    last_count <- server$ephemeral_counts[[key]]
    if (!is.null(last_count) && count <= last_count) {
      return(invisible())
    }
    server$ephemeral_counts[[key]] <- count
  }

  target_id <- msg$targetId
  doc_id <- msg$documentId

  if (!is.null(target_id) && exists(target_id, envir = server$connections)) {
    send_to_peer(server, target_id, msg)
  } else if (!is.null(doc_id)) {
    broadcast_ephemeral(server, client_id, doc_id, msg)
  }
  invisible()
}

#' Add a client to a document's peer list
#'
#' @param server An autosync_server object.
#' @param doc_id Document ID.
#' @param client_id Client's peer ID.
#'
#' @noRd
add_doc_peer <- function(server, doc_id, client_id) {
  peers <- server$doc_peers[[doc_id]]
  if (is.null(peers)) {
    server$doc_peers[[doc_id]] <- client_id
  } else if (!client_id %in% peers) {
    server$doc_peers[[doc_id]] <- c(peers, client_id)
  }
  invisible()
}

#' Remove a client from a document's peer list
#'
#' @param server An autosync_server object.
#' @param doc_id Document ID.
#' @param client_id Client's peer ID.
#'
#' @noRd
remove_doc_peer <- function(server, doc_id, client_id) {
  peers <- server$doc_peers[[doc_id]]
  if (!is.null(peers)) {
    peers <- peers[peers != client_id]
    if (length(peers) == 0L) {
      rm(list = doc_id, envir = server$doc_peers)
    } else {
      server$doc_peers[[doc_id]] <- peers
    }
  }
  invisible()
}

#' Remove a client from all document peer lists
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#'
#' @noRd
remove_peer_from_all_docs <- function(server, client_id) {
  for (doc_id in ls(server$doc_peers)) {
    remove_doc_peer(server, doc_id, client_id)
  }
  invisible()
}

#' Broadcast ephemeral message to all peers on a document
#'
#' @param server An autosync_server object.
#' @param sender_client_id ID of the client who sent the message.
#' @param doc_id Document ID.
#' @param msg The ephemeral message to broadcast.
#'
#' @noRd
broadcast_ephemeral <- function(server, sender_client_id, doc_id, msg) {
  peers <- server$doc_peers[[doc_id]]
  if (is.null(peers)) {
    return(invisible())
  }

  for (client_id in peers) {
    if (client_id == sender_client_id) {
      next
    }

    relay_msg <- msg
    relay_msg$targetId <- client_id
    send_to_peer(server, client_id, relay_msg)
  }

  invisible()
}

#' Handle error message from client
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#' @param msg Decoded error message.
#'
#' @noRd
handle_error <- function(server, client_id, msg) {
  warning("Error from client ", client_id, ": ", msg$message)
  invisible()
}

#' Clean up resources when client disconnects
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#'
#' @noRd
handle_disconnect <- function(server, client_id) {
  if (is.null(client_id)) {
    return(invisible())
  }
  if (exists(client_id, envir = server$sync_states)) {
    rm(list = client_id, envir = server$sync_states)
  }
  remove_peer_from_all_docs(server, client_id)

  # Clean up ephemeral count entries for this client
  prefix <- paste0(client_id, "\x1f")
  for (key in ls(server$ephemeral_counts)) {
    if (startsWith(key, prefix)) {
      rm(list = key, envir = server$ephemeral_counts)
    }
  }

  invisible()
}

#' Send message to specific peer
#'
#' @param server An autosync_server object.
#' @param peer_id Target peer ID.
#' @param msg List to encode and send.
#'
#' @noRd
send_to_peer <- function(server, peer_id, msg) {
  conn <- server$connections[[peer_id]]
  if (!is.null(conn) && !is.null(conn$ws)) {
    conn$ws$send(cborenc(msg))
  }
  invisible()
}

#' Send error message to peer
#'
#' @param server An autosync_server object.
#' @param peer_id Target peer ID.
#' @param message Error message string.
#' @param temp_id Optional temporary connection ID for pre-handshake errors.
#'
#' @noRd
send_error <- function(server, peer_id, message, temp_id = NULL) {
  response <- list(
    type = "error",
    senderId = server$peer_id,
    targetId = peer_id,
    message = message
  )
  if (!is.null(temp_id)) {
    conn <- server$connections[[temp_id]]
    if (!is.null(conn) && !is.null(conn$ws)) {
      conn$ws$send(cborenc(response))
    }
  } else {
    send_to_peer(server, peer_id, response)
  }
  invisible()
}

#' Send doc-unavailable message
#'
#' @param server An autosync_server object.
#' @param client_id Client's peer ID.
#' @param doc_id Document ID.
#'
#' @noRd
send_unavailable <- function(server, client_id, doc_id) {
  response <- list(
    type = "doc-unavailable",
    senderId = server$peer_id,
    targetId = client_id,
    documentId = doc_id
  )
  send_to_peer(server, client_id, response)
  invisible()
}

#' Announce a new document to all connected peers
#'
#' Sends an initial sync message for a single document to every peer connection.
#' Called when a document is created via [create_document()] after peering is
#' already established.
#'
#' @param server An autosync_server object.
#' @param doc_id Document ID.
#' @param doc Automerge document object.
#'
#' @noRd
announce_new_document <- function(server, doc_id, doc) {
  for (client_id in ls(server$connections)) {
    conn <- server$connections[[client_id]]
    if (is.null(conn$client_id) || client_id != conn$client_id) next
    if (!isTRUE(eval_share(server$share, client_id, doc_id))) next
    sync_to_peer(server, client_id, doc_id, doc)
  }
  invisible()
}

#' Broadcast sync changes to all connected peers
#'
#' @param server An autosync_server object.
#' @param sender_client_id ID of the client who sent the change.
#' @param doc_id Document ID.
#' @param doc Automerge document object.
#'
#' @noRd
broadcast_sync <- function(server, sender_client_id, doc_id, doc) {
  peers <- server$doc_peers[[doc_id]]
  if (is.null(peers)) {
    return(invisible())
  }

  for (client_id in peers) {
    if (client_id == sender_client_id) {
      next
    }

    sync_state <- server$sync_states[[client_id]][[doc_id]]

    reply_data <- am_sync_encode(doc, sync_state)
    if (!is.null(reply_data)) {
      response <- list(
        type = "sync",
        senderId = server$peer_id,
        targetId = client_id,
        documentId = doc_id,
        data = reply_data
      )
      send_to_peer(server, client_id, response)
    }
  }

  invisible()
}

# Test handlers by creating the state environment directly
# without using actual WebSocket connections

create_test_state <- function(data_dir = tempfile(), auto_create_docs = TRUE) {
  dir.create(data_dir, showWarnings = FALSE)

  state <- new.env(hash = TRUE, parent = emptyenv())
  state$data_dir <- data_dir
  state$auto_create_docs <- auto_create_docs
  state$peer_id <- autosync:::generate_peer_id()
  state$storage_id <- autosync:::generate_peer_id()
  state$documents <- new.env(hash = TRUE, parent = emptyenv())
  state$sync_states <- new.env(hash = TRUE, parent = emptyenv())
  state$connections <- new.env(hash = TRUE, parent = emptyenv())
  state$doc_peers <- new.env(hash = TRUE, parent = emptyenv())
  state$ephemeral_counts <- new.env(hash = TRUE, parent = emptyenv())
  state$share <- NA
  state
}

# Mock WebSocket object for testing
create_mock_ws <- function() {

  ws <- new.env(hash = TRUE)
  ws$sent_messages <- list()
  ws$closed <- FALSE
  ws$send <- function(data) {
    ws$sent_messages <- c(ws$sent_messages, list(data))
  }
  ws$close <- function() ws$closed <- TRUE
  ws$id <- as.character(sample.int(1e6, 1))
  ws
}

# Helper to set up nested sync state (sync_states[[client_id]][[doc_id]])
set_sync_state <- function(state, client_id, doc_id, sync_state = NULL) {
  if (is.null(state$sync_states[[client_id]])) {
    state$sync_states[[client_id]] <- new.env(hash = TRUE, parent = emptyenv())
  }
  state$sync_states[[client_id]][[doc_id]] <- sync_state %||%
    automerge::am_sync_state()
}

test_that("handle_message dispatches to handle_join", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Create mock connection
  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- secretbase::cborenc(list(
    type = "join",
    senderId = "testClient123",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  ))

  autosync:::handle_message(state, temp_id, temp_id, join_msg)

  # Should have sent a response
  expect_length(ws$sent_messages, 1)

  # Decode the response
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
  expect_equal(response$targetId, "testClient123")
  expect_equal(response$selectedProtocolVersion, "1")
})

test_that("handle_join rejects unsupported protocol version", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "badVersionClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("999") # Unsupported
  )

  autosync:::handle_join(state, temp_id, join_msg)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_true(grepl("protocol", response$message, ignore.case = TRUE))
})

test_that("handle_join sets up connection state correctly", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "setupClient",
    peerMetadata = list(customField = "value"),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  # Check connection was updated
  expect_equal(state$connections[[temp_id]]$client_id, "setupClient")
  expect_equal(state$connections[[temp_id]]$metadata$customField, "value")

  # Check client_id index was created
  expect_true(exists("setupClient", envir = state$connections))
})

test_that("handle_sync validates document ID format", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Set up connected client
  ws <- create_mock_ws()
  client_id <- "syncClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = "!!!invalid!!!", # Invalid base58check
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_true(grepl("Invalid", response$message, ignore.case = TRUE))
})

test_that("handle_sync sends doc-unavailable when auto_create_docs is FALSE", {
  state <- create_test_state(auto_create_docs = FALSE)
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  client_id <- "unavailableClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  doc_id <- generate_document_id()
  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = doc_id,
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "doc-unavailable")
  expect_equal(response$documentId, doc_id)
})

test_that("handle_sync auto-creates document when enabled", {
  state <- create_test_state(auto_create_docs = TRUE)
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  client_id <- "autoCreateClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  doc_id <- generate_document_id()
  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = doc_id,
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  # Document should be created
  expect_true(exists(doc_id, envir = state$documents))
  expect_true(inherits(state$documents[[doc_id]], "am_doc"))

  # Should have sent sync response
  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "sync")
  expect_equal(response$documentId, doc_id)
})

test_that("handle_sync creates sync state and doc_peers entry", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Pre-create a document
  doc <- automerge::am_create()
  doc_id <- generate_document_id()
  state$documents[[doc_id]] <- doc

  ws <- create_mock_ws()
  client_id <- "syncStateClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = doc_id,
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  # Sync state should be created (nested: sync_states[[client_id]][[doc_id]])
  expect_true(exists(client_id, envir = state$sync_states))
  expect_true(exists(doc_id, envir = state$sync_states[[client_id]]))

  # Client should be added to doc_peers
  expect_true(client_id %in% state$doc_peers[[doc_id]])
})

test_that("handle_sync ignores messages with wrong targetId", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  client_id <- "wrongTargetClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = "someOtherPeer", # Not the server's peer_id
    documentId = generate_document_id(),
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  # No response should be sent
  expect_length(ws$sent_messages, 0)
})

test_that("handle_message handles malformed CBOR gracefully", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  invalid_cbor <- as.raw(c(0xFF, 0xFF, 0xFF, 0xFF))

  # Should warn but not error
  expect_warning(
    autosync:::handle_message(state, temp_id, temp_id, invalid_cbor),
    "CBOR decode error"
  )
})

test_that("handle_message handles unknown message types", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = temp_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  unknown_msg <- secretbase::cborenc(list(
    type = "unknownType123",
    senderId = temp_id
  ))

  # Should warn but not error
  expect_warning(
    autosync:::handle_message(state, temp_id, temp_id, unknown_msg),
    "Unknown message type"
  )
})

test_that("handle_disconnect cleans up sync states and doc_peers", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  client_id <- "disconnectClient"
  doc_id <- generate_document_id()
  doc_id2 <- "anotherdoc"

  # Create sync states for this client (nested structure)
  set_sync_state(state, client_id, doc_id)
  set_sync_state(state, client_id, doc_id2)
  state$doc_peers[[doc_id]] <- c(client_id, "otherClient")
  state$doc_peers[[doc_id2]] <- client_id

  # Also create sync state for another client (should not be removed)
  set_sync_state(state, "otherClient", doc_id)

  autosync:::handle_disconnect(state, client_id)

  # Client's entire sync state env should be removed
  expect_false(exists(client_id, envir = state$sync_states))

  # Other client's sync state should remain
  expect_true(exists("otherClient", envir = state$sync_states))
  expect_true(exists(doc_id, envir = state$sync_states[["otherClient"]]))

  # Client should be removed from doc_peers
  expect_equal(state$doc_peers[[doc_id]], "otherClient")
  expect_false(exists(doc_id2, envir = state$doc_peers)) # empty list removed
})

test_that("handle_disconnect handles NULL client_id", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Should not error
  expect_no_error(autosync:::handle_disconnect(state, NULL))
})

test_that("handle_leave does nothing", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Should return invisibly without error
  result <- autosync:::handle_leave(state, "someClient", list(type = "leave"))
  expect_null(result)
})

test_that("handle_error logs warning", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  error_msg <- list(
    type = "error",
    senderId = "errorClient",
    message = "Test error message"
  )

  expect_warning(
    autosync:::handle_error(state, "errorClient", error_msg),
    "Test error message"
  )
})

test_that("handle_message dispatches leave, ephemeral, and error messages", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  client_id <- "dispatchClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  leave <- secretbase::cborenc(list(type = "leave", senderId = client_id))
  expect_no_error(autosync:::handle_message(state, client_id, client_id, leave))

  # An ephemeral broadcast to a document with no peers exits quietly.
  ephemeral <- secretbase::cborenc(list(
    type = "ephemeral",
    senderId = client_id,
    documentId = generate_document_id(),
    data = raw(0)
  ))
  expect_no_error(autosync:::handle_message(state, client_id, client_id, ephemeral))

  error <- secretbase::cborenc(list(
    type = "error",
    senderId = client_id,
    message = "boom"
  ))
  expect_warning(
    autosync:::handle_message(state, client_id, client_id, error),
    "boom"
  )
})

test_that("handle_join loads persisted sync states for a peer with a storageId", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Persist a sync state on disk for a storage peer.
  storage_id <- "joinStoragePeer"
  doc_id <- generate_document_id()
  autosync:::save_sync_state(state, storage_id, doc_id, automerge::am_sync_state())

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "storedClient",
    peerMetadata = list(storageId = storage_id),
    supportedProtocolVersions = list("1")
  )
  autosync:::handle_join(state, temp_id, join_msg)

  # The persisted sync state is loaded for this client.
  expect_true(exists("storedClient", envir = state$sync_states))
  expect_true(exists(doc_id, envir = state$sync_states[["storedClient"]]))
})

test_that("handle_sync warns on a sync decode error", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc_id <- generate_document_id()
  state$documents[[doc_id]] <- automerge::am_create()

  ws <- create_mock_ws()
  client_id <- "decodeErrClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = doc_id,
    data = as.raw(c(0xFF, 0xFF, 0xFF, 0xFF))
  )

  expect_warning(
    autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE),
    "am_sync_decode error"
  )
})

test_that("handle_sync persists sync state for a peer with a storageId", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc_id <- generate_document_id()
  state$documents[[doc_id]] <- automerge::am_create()

  ws <- create_mock_ws()
  client_id <- "storeSyncClient"
  storage_id <- "syncStore"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(storageId = storage_id),
    connected_at = Sys.time()
  )

  # Real sync data from a peer document so the change applies and persists.
  peer <- automerge::am_create()
  automerge::am_put(peer, automerge::AM_ROOT, "k", "v")
  data <- automerge::am_sync_encode(peer, automerge::am_sync_state())

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = doc_id,
    data = data
  )
  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  expect_true(file.exists(
    file.path(state$data_dir, ".sync_states", storage_id, paste0(doc_id, ".sync"))
  ))
})

test_that("handle_ephemeral forwards to target peer", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Set up two clients
  ws1 <- create_mock_ws()
  ws2 <- create_mock_ws()

  state$connections[["client1"]] <- list(
    ws = ws1,
    client_id = "client1",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["client2"]] <- list(
    ws = ws2,
    client_id = "client2",
    metadata = list(),
    connected_at = Sys.time()
  )

  ephemeral_msg <- list(
    type = "ephemeral",
    senderId = "client1",
    targetId = "client2",
    data = "hello"
  )

  autosync:::handle_ephemeral(state, "client1", ephemeral_msg)

  # client2 should receive the message
  expect_length(ws2$sent_messages, 1)
  forwarded <- secretbase::cbordec(ws2$sent_messages[[1]])
  expect_equal(forwarded$type, "ephemeral")
  expect_equal(forwarded$senderId, "client1")
  expect_equal(forwarded$data, "hello")
})

test_that("handle_ephemeral ignores message if target not connected", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws1 <- create_mock_ws()
  state$connections[["client1"]] <- list(
    ws = ws1,
    client_id = "client1",
    metadata = list(),
    connected_at = Sys.time()
  )

  ephemeral_msg <- list(
    type = "ephemeral",
    senderId = "client1",
    targetId = "nonexistentClient", # Not connected
    data = "hello"
  )

  # Should not error
  expect_no_error(
    autosync:::handle_ephemeral(state, "client1", ephemeral_msg)
  )
})

test_that("handle_ephemeral broadcasts to all peers on document", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Set up three clients
  ws1 <- create_mock_ws()
  ws2 <- create_mock_ws()
  ws3 <- create_mock_ws()

  state$connections[["client1"]] <- list(
    ws = ws1,
    client_id = "client1",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["client2"]] <- list(
    ws = ws2,
    client_id = "client2",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["client3"]] <- list(
    ws = ws3,
    client_id = "client3",
    metadata = list(),
    connected_at = Sys.time()
  )

  # Create a document and sync states for clients 1 and 2 (but not 3)
  doc_id <- "testDoc123"
  state$documents[[doc_id]] <- automerge::am_create()
  set_sync_state(state, "client1", doc_id)
  set_sync_state(state, "client2", doc_id)
  state$doc_peers[[doc_id]] <- c("client1", "client2")
  # client3 has no sync state for this document

  # Broadcast ephemeral message from client1 (has documentId, no targetId)
  ephemeral_msg <- list(
    type = "ephemeral",
    senderId = "client1",
    documentId = doc_id,
    data = list(presence = "cursor at line 5")
  )

  autosync:::handle_ephemeral(state, "client1", ephemeral_msg)

  # client1 (sender) should NOT receive the message
  expect_length(ws1$sent_messages, 0)

  # client2 (has sync state) should receive the message
  expect_length(ws2$sent_messages, 1)
  forwarded <- secretbase::cbordec(ws2$sent_messages[[1]])
  expect_equal(forwarded$type, "ephemeral")
  expect_equal(forwarded$senderId, "client1")
  expect_equal(forwarded$documentId, doc_id)
  expect_equal(forwarded$targetId, "client2")
  expect_equal(forwarded$data$presence, "cursor at line 5")

  # client3 (no sync state) should NOT receive the message
  expect_length(ws3$sent_messages, 0)
})

test_that("handle_ephemeral prefers targetId over documentId broadcast", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Set up two clients with sync state
  ws1 <- create_mock_ws()
  ws2 <- create_mock_ws()
  ws3 <- create_mock_ws()

  state$connections[["client1"]] <- list(
    ws = ws1,
    client_id = "client1",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["client2"]] <- list(
    ws = ws2,
    client_id = "client2",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["client3"]] <- list(
    ws = ws3,
    client_id = "client3",
    metadata = list(),
    connected_at = Sys.time()
  )

  doc_id <- "testDoc456"
  state$documents[[doc_id]] <- automerge::am_create()
  set_sync_state(state, "client1", doc_id)
  set_sync_state(state, "client2", doc_id)
  set_sync_state(state, "client3", doc_id)
  state$doc_peers[[doc_id]] <- c("client1", "client2", "client3")

  # Message with BOTH targetId and documentId - should use point-to-point
  ephemeral_msg <- list(
    type = "ephemeral",
    senderId = "client1",
    targetId = "client2",
    documentId = doc_id,
    data = "direct message"
  )

  autosync:::handle_ephemeral(state, "client1", ephemeral_msg)

  # Only client2 should receive (point-to-point, not broadcast)
  expect_length(ws2$sent_messages, 1)
  expect_length(ws3$sent_messages, 0)
})

test_that("send_to_peer sends encoded message", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  peer_id <- "testPeer"
  state$connections[[peer_id]] <- list(
    ws = ws,
    client_id = peer_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  msg <- list(type = "test", data = "hello")
  autosync:::send_to_peer(state, peer_id, msg)

  expect_length(ws$sent_messages, 1)
  decoded <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(decoded$type, "test")
  expect_equal(decoded$data, "hello")
})

test_that("send_to_peer handles missing peer", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  msg <- list(type = "test")

  # Should not error when peer doesn't exist
  expect_no_error(
    autosync:::send_to_peer(state, "nonexistent", msg)
  )
})

test_that("send_error sends error message", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  peer_id <- "errorPeer"
  state$connections[[peer_id]] <- list(
    ws = ws,
    client_id = peer_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  autosync:::send_error(state, peer_id, "Something went wrong")

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_equal(response$targetId, peer_id)
  expect_equal(response$message, "Something went wrong")
})

test_that("send_error with temp_id sends to temp connection", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  autosync:::send_error(state, "targetPeer", "Error message", temp_id = temp_id)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_equal(response$targetId, "targetPeer")
})

test_that("send_unavailable sends doc-unavailable message", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  client_id <- "unavailablePeer"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  doc_id <- "missingDoc123"
  autosync:::send_unavailable(state, client_id, doc_id)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "doc-unavailable")
  expect_equal(response$targetId, client_id)
  expect_equal(response$documentId, doc_id)
})

test_that("broadcast_sync sends to other peers with sync state", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Create a document
  doc <- automerge::am_create()
  doc_id <- generate_document_id()
  state$documents[[doc_id]] <- doc

  # Set up multiple clients
  ws1 <- create_mock_ws()
  ws2 <- create_mock_ws()
  ws3 <- create_mock_ws()

  state$connections[["sender"]] <- list(
    ws = ws1,
    client_id = "sender",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["receiver1"]] <- list(
    ws = ws2,
    client_id = "receiver1",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["receiver2"]] <- list(
    ws = ws3,
    client_id = "receiver2",
    metadata = list(),
    connected_at = Sys.time()
  )

  # Create sync states for receivers (not sender)
  set_sync_state(state, "receiver1", doc_id)
  set_sync_state(state, "receiver2", doc_id)
  state$doc_peers[[doc_id]] <- c("receiver1", "receiver2")

  autosync:::broadcast_sync(state, "sender", doc_id, doc)

  # Sender should not receive
  expect_length(ws1$sent_messages, 0)

  # Receivers should receive sync messages (or nothing if no new data)
  # The actual content depends on sync state
})

test_that("broadcast_sync skips pre-handshake connections", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc <- automerge::am_create()
  doc_id <- generate_document_id()
  state$documents[[doc_id]] <- doc

  ws <- create_mock_ws()
  temp_id <- ws$id

  # Pre-handshake connection (client_id is NULL)
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  autosync:::broadcast_sync(state, "sender", doc_id, doc)

  # Should not send to pre-handshake connection
  expect_length(ws$sent_messages, 0)
})

test_that("handle_join with ephemeral server sets isEphemeral=TRUE in response", {
  state <- create_test_state()
  state$storage_id <- NULL # Ephemeral server
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "ephemeralTestClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
  expect_true(response$peerMetadata$isEphemeral)
  expect_null(response$peerMetadata$storageId)
})

test_that("handle_join with persistent server includes storageId", {
  state <- create_test_state()
  state$storage_id <- "myStorageId123"
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "persistentTestClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
  expect_false(response$peerMetadata$isEphemeral)
  expect_equal(response$peerMetadata$storageId, "myStorageId123")
})

test_that("handle_join accepts missing supportedProtocolVersions", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "noVersionClient",
    peerMetadata = list(isEphemeral = TRUE)
    # supportedProtocolVersions intentionally omitted
  )

  autosync:::handle_join(state, temp_id, join_msg)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
  expect_equal(response$targetId, "noVersionClient")
  expect_equal(response$selectedProtocolVersion, "1")
})

test_that("handle_join closes previous socket on duplicate senderId", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # First connection
  ws1 <- create_mock_ws()
  temp_id1 <- ws1$id
  state$connections[[temp_id1]] <- list(
    ws = ws1,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg1 <- list(
    type = "join",
    senderId = "reconnectClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id1, join_msg1)

  # Verify first connection is active
  expect_length(ws1$sent_messages, 1)
  expect_true(exists("reconnectClient", envir = state$connections))

  # Second connection with same senderId

  ws2 <- create_mock_ws()
  temp_id2 <- ws2$id
  state$connections[[temp_id2]] <- list(
    ws = ws2,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg2 <- list(
    type = "join",
    senderId = "reconnectClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id2, join_msg2)

  # Old WS should have been closed
  expect_true(ws1$closed)

  # New connection should be active
  expect_length(ws2$sent_messages, 1)
  response <- secretbase::cbordec(ws2$sent_messages[[1]])
  expect_equal(response$type, "peer")
  expect_equal(response$targetId, "reconnectClient")

  # Connection env should point to new ws
  conn <- state$connections[["reconnectClient"]]
  expect_identical(conn$ws, ws2)
})

test_that("handle_join with share=TRUE announces all documents", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- TRUE

  # Pre-create two documents with data
  doc1 <- automerge::am_create()
  automerge::am_put(doc1, automerge::AM_ROOT, "key1", "value1")
  doc_id1 <- generate_document_id()
  state$documents[[doc_id1]] <- doc1

  doc2 <- automerge::am_create()
  automerge::am_put(doc2, automerge::AM_ROOT, "key2", "value2")
  doc_id2 <- generate_document_id()
  state$documents[[doc_id2]] <- doc2

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "peerClient",
    peerMetadata = list(isEphemeral = FALSE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  # Should receive: 1 peer response + 2 sync messages (one per document)
  expect_length(ws$sent_messages, 3)

  types <- vapply(ws$sent_messages, function(m) secretbase::cbordec(m)$type, "")
  expect_equal(types[1], "peer")
  expect_equal(sort(types[2:3]), c("sync", "sync"))

  doc_ids <- vapply(ws$sent_messages[2:3], function(m) {
    secretbase::cbordec(m)$documentId
  }, "")
  expect_true(doc_id1 %in% doc_ids)
  expect_true(doc_id2 %in% doc_ids)

  # Sync states and doc_peers should be set up
  expect_true(exists("peerClient", envir = state$sync_states))
  expect_true("peerClient" %in% state$doc_peers[[doc_id1]])
  expect_true("peerClient" %in% state$doc_peers[[doc_id2]])
})

test_that("handle_join with share=NA does not announce documents", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  # Pre-create a document
  doc <- automerge::am_create()
  automerge::am_put(doc, automerge::AM_ROOT, "key", "value")
  doc_id <- generate_document_id()
  state$documents[[doc_id]] <- doc

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "regularClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  # Should only receive the peer response, no sync messages
  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
})

test_that("broadcast_ephemeral preserves count and sessionId from sender", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws1 <- create_mock_ws()
  ws2 <- create_mock_ws()

  state$connections[["sender"]] <- list(
    ws = ws1,
    client_id = "sender",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["receiver"]] <- list(
    ws = ws2,
    client_id = "receiver",
    metadata = list(),
    connected_at = Sys.time()
  )

  doc_id <- "ephTestDoc"
  state$doc_peers[[doc_id]] <- c("sender", "receiver")

  ephemeral_msg <- list(
    type = "ephemeral",
    senderId = "sender",
    documentId = doc_id,
    count = 42L,
    sessionId = "session-abc-123",
    data = list(cursor = "line 10")
  )

  autosync:::handle_ephemeral(state, "sender", ephemeral_msg)

  expect_length(ws2$sent_messages, 1)
  relayed <- secretbase::cbordec(ws2$sent_messages[[1]])
  expect_equal(relayed$type, "ephemeral")
  expect_equal(relayed$count, 42L)
  expect_equal(relayed$sessionId, "session-abc-123")
  expect_equal(relayed$senderId, "sender")
  expect_equal(relayed$data$cursor, "line 10")
})

# --- RFC compliance tests ---

test_that("handle_join accepts version list containing '1' among others", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "multiVersionClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1", "2")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
  expect_equal(response$selectedProtocolVersion, "1")
})

test_that("handle_join rejects version list not containing '1'", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "futureClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("2")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_true(grepl("protocol", response$message, ignore.case = TRUE))
})

test_that("handle_join closes connection after protocol version error", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "badVersionClient2",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("999")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  expect_true(ws$closed)
})

test_that("handle_sync closes connection after invalid document ID error", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  client_id <- "invalidDocClient"
  state$connections[[client_id]] <- list(
    ws = ws,
    client_id = client_id,
    metadata = list(),
    connected_at = Sys.time()
  )

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = "!!!invalid!!!",
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_true(ws$closed)
})

test_that("handle_message rejects non-join messages before handshake", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  sync_msg <- secretbase::cborenc(list(
    type = "sync",
    senderId = "sneakyClient",
    documentId = "someDoc",
    data = raw(0)
  ))

  autosync:::handle_message(state, temp_id, temp_id, sync_msg)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "error")
  expect_true(grepl("join", response$message, ignore.case = TRUE))
  expect_true(ws$closed)
})

test_that("handle_message discards zero-length messages", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws,
    client_id = NULL,
    metadata = NULL,
    connected_at = Sys.time()
  )

  expect_no_error(
    autosync:::handle_message(state, temp_id, temp_id, raw())
  )
  expect_length(ws$sent_messages, 0)
})

test_that("handle_ephemeral deduplicates by count and sessionId", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  ws1 <- create_mock_ws()
  ws2 <- create_mock_ws()

  state$connections[["sender"]] <- list(
    ws = ws1,
    client_id = "sender",
    metadata = list(),
    connected_at = Sys.time()
  )

  state$connections[["receiver"]] <- list(
    ws = ws2,
    client_id = "receiver",
    metadata = list(),
    connected_at = Sys.time()
  )

  doc_id <- "dedupDoc"
  state$doc_peers[[doc_id]] <- c("sender", "receiver")

  # First message with count = 2 should be forwarded
  msg1 <- list(
    type = "ephemeral",
    senderId = "sender",
    documentId = doc_id,
    sessionId = "session-1",
    count = 2L,
    data = "first"
  )
  autosync:::handle_ephemeral(state, "sender", msg1)
  expect_length(ws2$sent_messages, 1)

  # Second message with count = 1 (stale) should be discarded
  msg2 <- list(
    type = "ephemeral",
    senderId = "sender",
    documentId = doc_id,
    sessionId = "session-1",
    count = 1L,
    data = "stale"
  )
  autosync:::handle_ephemeral(state, "sender", msg2)
  expect_length(ws2$sent_messages, 1) # Still 1, not forwarded

  # Third message with count = 2 (equal) should also be discarded
  msg3 <- list(
    type = "ephemeral",
    senderId = "sender",
    documentId = doc_id,
    sessionId = "session-1",
    count = 2L,
    data = "duplicate"
  )
  autosync:::handle_ephemeral(state, "sender", msg3)
  expect_length(ws2$sent_messages, 1) # Still 1

  # Fourth message with count = 3 should be forwarded
  msg4 <- list(
    type = "ephemeral",
    senderId = "sender",
    documentId = doc_id,
    sessionId = "session-1",
    count = 3L,
    data = "new"
  )
  autosync:::handle_ephemeral(state, "sender", msg4)
  expect_length(ws2$sent_messages, 2)
})

test_that("handle_disconnect cleans up ephemeral counts", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  client_id <- "ephCleanupClient"
  state$ephemeral_counts[[paste0(client_id, "\x1f", "session1")]] <- 5L
  state$ephemeral_counts[[paste0(client_id, "\x1f", "session2")]] <- 3L
  state$ephemeral_counts[[paste0("otherClient", "\x1f", "session1")]] <- 7L

  autosync:::handle_disconnect(state, client_id)

  expect_equal(length(ls(state$ephemeral_counts)), 1)
  expect_equal(state$ephemeral_counts[[paste0("otherClient", "\x1f", "session1")]], 7L)
})

# --- share policy tests ---

test_that("share=TRUE announces documents to all peers on join", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- TRUE

  doc <- automerge::am_create()
  automerge::am_put(doc, automerge::AM_ROOT, "key", "value")
  doc_id <- autosync:::generate_document_id()
  state$documents[[doc_id]] <- doc

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws, client_id = NULL, metadata = NULL, connected_at = Sys.time()
  )

  # Regular client (no isPeer) should still get announcements with share=TRUE
  join_msg <- list(
    type = "join",
    senderId = "shareClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  # peer response + sync announcement
  expect_length(ws$sent_messages, 2)
  types <- vapply(ws$sent_messages, function(m) secretbase::cbordec(m)$type, "")
  expect_equal(types, c("peer", "sync"))
})

test_that("share=FALSE denies document requests", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- FALSE

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()
  state$documents[[doc_id]] <- doc

  ws <- create_mock_ws()
  client_id <- "deniedClient"
  state$connections[[client_id]] <- list(
    ws = ws, client_id = client_id, metadata = list(), connected_at = Sys.time()
  )

  sync_msg <- list(
    type = "request",
    senderId = client_id,
    targetId = state$peer_id,
    documentId = doc_id,
    data = raw(0)
  )

  autosync:::handle_sync(state, client_id, sync_msg, is_request = TRUE)

  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "doc-unavailable")
  expect_equal(response$documentId, doc_id)
})

test_that("share=FALSE does not announce on join", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- FALSE

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()
  state$documents[[doc_id]] <- doc

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws, client_id = NULL, metadata = NULL, connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "deniedPeer",
    peerMetadata = list(isPeer = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  # Only peer response, no sync announcements
  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "peer")
})

test_that("share=NA allows requests but does not announce", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- NA

  doc <- automerge::am_create()
  automerge::am_put(doc, automerge::AM_ROOT, "key", "value")
  doc_id <- autosync:::generate_document_id()
  state$documents[[doc_id]] <- doc

  # Test no announce on join
  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws, client_id = NULL, metadata = NULL, connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "naClient",
    peerMetadata = list(isPeer = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)
  expect_length(ws$sent_messages, 1)  # peer response only

  # Test request is allowed
  sync_msg <- list(
    type = "request",
    senderId = "naClient",
    targetId = state$peer_id,
    documentId = doc_id,
    data = raw(0)
  )

  autosync:::handle_sync(state, "naClient", sync_msg, is_request = TRUE)

  # Should get a sync response (not doc-unavailable)
  expect_length(ws$sent_messages, 2)
  response <- secretbase::cbordec(ws$sent_messages[[2]])
  expect_equal(response$type, "sync")
  expect_equal(response$documentId, doc_id)
})

test_that("share function controls announce per peer and document", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc1 <- automerge::am_create()
  doc_id1 <- autosync:::generate_document_id()
  state$documents[[doc_id1]] <- doc1

  doc2 <- automerge::am_create()
  doc_id2 <- autosync:::generate_document_id()
  state$documents[[doc_id2]] <- doc2

  # Only announce doc1
  state$share <- function(client_id, doc_id) {
    if (doc_id == doc_id1) TRUE else NA
  }

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws, client_id = NULL, metadata = NULL, connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "selectiveClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  # peer response + 1 sync (only doc1 announced)
  expect_length(ws$sent_messages, 2)
  types <- vapply(ws$sent_messages, function(m) secretbase::cbordec(m)$type, "")
  expect_equal(types, c("peer", "sync"))
  sync_msg <- secretbase::cbordec(ws$sent_messages[[2]])
  expect_equal(sync_msg$documentId, doc_id1)
})

test_that("share function controls access per peer and document", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc1 <- automerge::am_create()
  doc_id1 <- autosync:::generate_document_id()
  state$documents[[doc_id1]] <- doc1

  doc2 <- automerge::am_create()
  doc_id2 <- autosync:::generate_document_id()
  state$documents[[doc_id2]] <- doc2

  # Allow doc1, deny doc2
  state$share <- function(client_id, doc_id) {
    if (doc_id == doc_id1) NA else FALSE
  }

  ws <- create_mock_ws()
  client_id <- "aclClient"
  state$connections[[client_id]] <- list(
    ws = ws, client_id = client_id, metadata = list(), connected_at = Sys.time()
  )

  # Request doc1 — should succeed
  msg1 <- list(
    type = "request", senderId = client_id,
    targetId = state$peer_id, documentId = doc_id1, data = raw(0)
  )
  autosync:::handle_sync(state, client_id, msg1, is_request = TRUE)

  expect_length(ws$sent_messages, 1)
  resp1 <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(resp1$type, "sync")

  # Request doc2 — should be denied
  msg2 <- list(
    type = "request", senderId = client_id,
    targetId = state$peer_id, documentId = doc_id2, data = raw(0)
  )
  autosync:::handle_sync(state, client_id, msg2, is_request = TRUE)

  expect_length(ws$sent_messages, 2)
  resp2 <- secretbase::cbordec(ws$sent_messages[[2]])
  expect_equal(resp2$type, "doc-unavailable")
  expect_equal(resp2$documentId, doc_id2)
})

test_that("share function receives client_id", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()
  state$documents[[doc_id]] <- doc

  captured_id <- NULL
  state$share <- function(client_id, doc_id) {
    captured_id <<- client_id
    TRUE
  }

  ws <- create_mock_ws()
  temp_id <- ws$id
  state$connections[[temp_id]] <- list(
    ws = ws, client_id = NULL, metadata = NULL, connected_at = Sys.time()
  )

  join_msg <- list(
    type = "join",
    senderId = "metadataClient",
    peerMetadata = list(isEphemeral = TRUE),
    supportedProtocolVersions = list("1")
  )

  autosync:::handle_join(state, temp_id, join_msg)

  expect_equal(captured_id, "metadataClient")
})

test_that("announce_new_document respects share policy", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- TRUE

  ws <- create_mock_ws()
  client_id <- "announceShareClient"
  state$connections[[client_id]] <- list(
    ws = ws, client_id = client_id,
    metadata = list(isEphemeral = TRUE),
    connected_at = Sys.time()
  )

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()

  autosync:::announce_new_document(state, doc_id, doc)

  # share=TRUE means announce to all connected peers
  expect_length(ws$sent_messages, 1)
  response <- secretbase::cbordec(ws$sent_messages[[1]])
  expect_equal(response$type, "sync")
  expect_equal(response$documentId, doc_id)
})

test_that("announce_new_document skips connections when share=NA", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()

  client_id <- "regularClient"
  ws <- create_mock_ws()
  state$connections[[client_id]] <- list(
    ws = ws, client_id = client_id,
    metadata = list(), connected_at = Sys.time()
  )

  autosync:::announce_new_document(state, doc_id, doc)

  expect_length(ws$sent_messages, 0)
})

test_that("announce_new_document skips pre-handshake connections", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- TRUE

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()

  temp_id <- "tempWsId"
  ws <- create_mock_ws()
  state$connections[[temp_id]] <- list(
    ws = ws, client_id = NULL,
    metadata = NULL, connected_at = Sys.time()
  )

  autosync:::announce_new_document(state, doc_id, doc)

  expect_length(ws$sent_messages, 0)
})

test_that("announce_new_document skips duplicate temp_id entries", {
  state <- create_test_state()
  on.exit(unlink(state$data_dir, recursive = TRUE))
  state$share <- TRUE

  doc <- automerge::am_create()
  doc_id <- autosync:::generate_document_id()

  peer_id <- "dualIndexPeer"
  temp_id <- "tempId123"
  ws <- create_mock_ws()
  conn <- list(
    ws = ws, client_id = peer_id,
    metadata = list(), connected_at = Sys.time()
  )
  # Both entries point to the same connection (dual indexing)
  state$connections[[peer_id]] <- conn
  state$connections[[temp_id]] <- conn

  autosync:::announce_new_document(state, doc_id, doc)

  # Should send exactly once (skip temp_id entry where key != client_id)
  expect_length(ws$sent_messages, 1)
})

# Test helpers

# Drain any leftover later callbacks from prior tests so they don't fire
# inside the next test's run_now and trip up nanonext's external-pointer
# bookkeeping. Bounded so a perpetually-rescheduling callback can't hang.
drain_later <- function(max_iters = 10L) {
  tryCatch(
    for (i in seq_len(max_iters)) {
      if (later::loop_empty()) {
        break
      }
      later::run_now(0.1)
    },
    error = function(e) NULL
  )
}

# auth_config validates issuer and client_id

    Code
      auth_config(issuer = 123, client_id = "x")
    Condition
      Error in `auth_config()`:
      ! 'issuer' must be a single character string (OIDC issuer URL)

---

    Code
      auth_config(issuer = "x", client_id = 123)
    Condition
      Error in `auth_config()`:
      ! 'client_id' must be set (or set the OIDC_CLIENT_ID environment variable)

---

    Code
      auth_config(client_id = "")
    Condition
      Error in `auth_config()`:
      ! 'client_id' must be set (or set the OIDC_CLIENT_ID environment variable)

# server requires TLS when auth is enabled

    Code
      sync_server(auth = auth_config(issuer = "https://accounts.google.com",
        client_id = "test-id"))
    Condition
      Error in `sync_server()`:
      ! Authentication requires TLS. Provide a 'tls' configuration.
      Transmitting tokens over unencrypted connections is a security risk.

# server rejects unauthenticated client when auth enabled

    Code
      sync_fetch(url = server$url, doc_id = generate_document_id(), tls = client_tls)
    Condition
      Error in `sync_fetch()`:
      ! Server error: Authentication failed

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{get_signing_key}
\alias{get_signing_key}
\title{Get a signing key for JWT verification}
\usage{
get_signing_key(issuer, kid)
}
\arguments{
\item{issuer}{The OIDC issuer URL.}

\item{kid}{The key ID from the JWT header.}
}
\value{
A public key object, or NULL if not found.
}
\description{
Looks up the public key for the given issuer and key ID, refreshing the
cache if needed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{fetch_jwks}
\alias{fetch_jwks}
\title{Fetch and parse a JWKS endpoint}
\usage{
fetch_jwks(jwks_uri)
}
\arguments{
\item{jwks_uri}{URL of the JWKS endpoint.}
}
\value{
List with \code{keys} (named list of jose key objects by kid) and
\code{expiry} (POSIXct when the cache should be refreshed).
}
\description{
Fetch and parse a JWKS endpoint
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{authenticate_header}
\alias{authenticate_header}
\title{Authenticate a client from HTTP request headers}
\usage{
authenticate_header(auth_config, headers)
}
\arguments{
\item{auth_config}{A \code{autosync_auth_config} object.}

\item{headers}{Named list of HTTP request headers.}
}
\value{
List with \code{valid} (logical), \code{email} (character or NULL),
\code{error} (character or NULL).
}
\description{
Extracts and validates a Bearer token (JWT) from the Authorization header
of the WebSocket upgrade request.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{generate_document_id}
\alias{generate_document_id}
\title{Generate a new document ID}
\usage{
generate_document_id()
}
\value{
Character string (Base58Check encoded).
}
\description{
Creates a new unique document ID compatible with automerge-repo.
The ID is a 16-byte random value encoded with Base58Check.
}
\examples{
generate_document_id()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{discover_jwks_uri}
\alias{discover_jwks_uri}
\title{Discover the JWKS URI for an OIDC issuer}
\usage{
discover_jwks_uri(issuer)
}
\arguments{
\item{issuer}{The OIDC issuer URL.}
}
\value{
Character string, the JWKS URI.
}
\description{
Discover the JWKS URI for an OIDC issuer
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/client.R
\name{sync_client}
\alias{sync_client}
\title{Open a persistent sync connection}
\usage{
sync_client(url, timeout = 5000L, tls = NULL, token = NULL, interval = 1000L)
}
\arguments{
\item{url}{WebSocket URL of the sync server (e.g., "ws://localhost:3030/"
or "wss://sync.automerge.org/"). Note: trailing slash may be required.}

\item{timeout}{Timeout in milliseconds for each receive operation. Default 5000.}

\item{tls}{(optional) for secure wss:// connections to servers with
self-signed or custom CA certificates, a TLS configuration object
created by \code{\link[nanonext:tls_config]{nanonext::tls_config()}}.}

\item{token}{(optional) JWT (ID token) for authenticated servers.
Sent as a Bearer token in the Authorization header of the WebSocket
upgrade request.}

\item{interval}{Interval in milliseconds for pushing local changes to
the server. Default 1000. Uses \code{\link[later:later]{later::later()}} to periodically check
for and send local changes for every open document. This is a cheap no-op
when there are no changes.}
}
\value{
An environment of class \code{"autosync_client"} with reference semantics,
representing the connection:
\describe{
\item{\code{open_doc(doc_id, timeout)}}{Open a live document over this
connection and return a \code{autosync_doc} handle for it (see below).
Repeated calls for the same \code{doc_id} reuse the document already open
on the connection rather than requesting it again.}
\item{\code{close()}}{Disconnect and stop syncing all open documents.}
\item{\code{active}}{Logical, whether the connection is active.}
}

A \code{autosync_doc} handle returned by \verb{$open_doc()} is itself an environment
with:
\describe{
\item{\code{doc}}{The live automerge document, kept in sync with the server.}
\item{\code{push()}}{Push this document's local changes to the server
immediately.}
\item{\code{close()}}{Stop syncing this one document (detach it from the
connection); the connection and its other documents are unaffected.}
\item{\code{active}}{Logical, whether the document is still open on an active
connection.}
}
}
\description{
Connects to an automerge-repo sync server and maintains a persistent
WebSocket connection. The connection performs the protocol handshake but
holds no documents on its own; open one or more live documents over it with
the \verb{$open_doc()} method. Each document stays synced — receiving real-time
updates from other peers and flushing local changes — for as long as the
connection is open. Unlike \code{\link[=sync_fetch]{sync_fetch()}}, which performs a one-off
retrieval over a throwaway connection, several documents can share a single
connection here.
}
\details{
Opening the connection performs a synchronous handshake before returning.
\verb{$open_doc()} then performs a synchronous initial sync, so the returned
handle's \verb{$doc} has meaningful content immediately. After that, incoming
changes are received asynchronously via a self-chaining promise loop, and
local changes are flushed periodically via a \code{\link[later:later]{later::later()}} timer.

Neither \code{close()} flushes pending local changes. Call \verb{$push()} first if you
have unsynced edits — otherwise any changes made since the last
\code{sync}-interval tick may be lost.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
server <- sync_server()
server$start()
doc_id <- create_document(server)

conn <- sync_client(server$url)
doc <- conn$open_doc(doc_id)
automerge::am_keys(doc$doc)

# Make local changes and push
automerge::am_put(doc$doc, automerge::AM_ROOT, "key", "value")
doc$push()

# Open another document over the same connection
other <- conn$open_doc(create_document(server))

# Disconnect (closes every document on the connection)
conn$close()
server$close()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/server.R
\name{list_documents}
\alias{list_documents}
\title{List all document IDs}
\usage{
list_documents(server)
}
\arguments{
\item{server}{An autosync_server object.}
}
\value{
Character vector of document IDs.
}
\description{
Returns the IDs of all documents currently loaded in the server.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
server <- sync_server()
create_document(server)
list_documents(server)
server$close()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/client.R
\name{sync_fetch}
\alias{sync_fetch}
\title{Fetch a document from a sync server}
\usage{
sync_fetch(
  url,
  doc_id,
  timeout = 5000L,
  tls = NULL,
  token = NULL,
  verbose = FALSE
)
}
\arguments{
\item{url}{WebSocket URL of the sync server (e.g., "ws://localhost:3030/"
or "wss://sync.automerge.org/"). Note: trailing slash may be required.}

\item{doc_id}{Document ID (base58check encoded string)}

\item{timeout}{Timeout in milliseconds for each receive operation. Default 5000.}

\item{tls}{(optional) for secure wss:// connections to servers with
self-signed or custom CA certificates, a TLS configuration object
created by \code{\link[nanonext:tls_config]{nanonext::tls_config()}}.}

\item{token}{(optional) JWT (ID token) for authenticated servers.
Sent as a Bearer token in the Authorization header of the WebSocket
upgrade request.}

\item{verbose}{Logical, print debug messages. Default FALSE.}
}
\value{
An automerge document object containing the fetched data.
}
\description{
Connects to an automerge-repo sync server and retrieves a document by ID.
This is useful for debugging sync issues or fetching documents from remote
servers like sync.automerge.org.
}
\details{
The function implements the automerge-repo sync protocol:
\enumerate{
\item Connects via WebSocket
\item Sends join message with peer ID
\item Receives peer response from server
\item Sends sync request for the specified document
\item Receives and applies sync messages until complete
}

Sync is considered complete when no new messages arrive within the timeout
after at least one sync round.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
# Fetch from public sync server
doc <- sync_fetch("wss://sync.automerge.org", "4F63WJPDzbHkkfKa66h1Qrr1sC5U")

# Fetch from local server with debug output
doc <- sync_fetch(server$url, "myDocId", verbose = TRUE)

# Fetch from server with self-signed certificate
cert <- nanonext::write_cert()
tls <- nanonext::tls_config(client = cert$client)
doc <- sync_fetch(server$url, "myDocId", tls = tls)

# Fetch from authenticated server
doc <- sync_fetch(
  "wss://secure.example.com",
  "myDocId",
  token = "eyJhbGciOi..."
)

# Inspect the document
automerge::am_keys(doc)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{sync_token}
\alias{sync_token}
\title{Obtain an OIDC token interactively}
\usage{
sync_token(
  client_id = Sys.getenv("OIDC_CLIENT_ID"),
  client_secret = Sys.getenv("OIDC_CLIENT_SECRET"),
  issuer = oidc_issuer(),
  scopes = "openid email",
  redirect_uri = oauth_redirect_uri()
)
}
\arguments{
\item{client_id}{The OIDC client ID (application ID). Defaults to the
\code{OIDC_CLIENT_ID} environment variable.}

\item{client_secret}{The OIDC client secret. Required by Google (Desktop
app) and "Web application" client types; leave unset for native / public
clients, which authenticate via PKCE alone. Defaults to the
\code{OIDC_CLIENT_SECRET} environment variable.}

\item{issuer}{The OIDC issuer URL. Defaults to the \code{OIDC_ISSUER}
environment variable, falling back to Google
(\code{"https://accounts.google.com"}).}

\item{scopes}{Space-separated OAuth scopes to request. Default
\code{"openid email"}.}

\item{redirect_uri}{Local redirect URI for the OAuth callback. Defaults to
\code{\link[httr2:oauth_redirect_uri]{httr2::oauth_redirect_uri()}}, i.e. \code{"http://localhost"} with an
OS-assigned random port (or the \code{HTTR2_OAUTH_REDIRECT_URL} environment
variable on hosted platforms). Supply an explicit port (e.g.
\code{"http://localhost:8080"}) when your OIDC provider requires the redirect
URI to match a pre-registered value.}
}
\value{
A JWT (ID token) as a character string.
}
\description{
Performs the OAuth 2.0 Authorization Code flow with PKCE to obtain a JWT
(ID token) from an OIDC provider, delegating the browser handshake and token
exchange to \pkg{httr2}. Endpoints are discovered from the issuer's
\code{.well-known} metadata via \code{\link[httr2:oauth_server_metadata]{httr2::oauth_server_metadata()}}, and the flow is
run by \code{\link[httr2:oauth_flow_auth_code]{httr2::oauth_flow_auth_code()}}: it opens the system browser, listens
on a loopback redirect for the callback, and returns the ID token for use
with \code{\link[=sync_fetch]{sync_fetch()}}.
}
\details{
For Google, register the OAuth client as a "Desktop app" and set both
\code{OIDC_CLIENT_ID} and \code{OIDC_CLIENT_SECRET}. Google's Desktop app secret is
required in the token exchange but, unlike a "Web application" secret, is
not treated as confidential: Google states that for installed apps "the
client secret is obviously not treated as a secret"
(\url{https://developers.google.com/identity/protocols/oauth2#installed}),
consistent with the OAuth 2.0 for Native Apps standard
(RFC 8252 section 8.5,
\url{https://datatracker.ietf.org/doc/html/rfc8252#section-8.5}).
Providers that support native / public clients (Microsoft Entra, Okta,
Auth0, etc.) need only \code{client_id}, authenticating via PKCE alone; leave
\code{client_secret} unset for these.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
# Uses OIDC_CLIENT_ID and OIDC_CLIENT_SECRET env vars by default
token <- sync_token()

# Or supply credentials directly
token <- sync_token(
  client_id = "YOUR_CLIENT_ID.apps.googleusercontent.com",
  client_secret = "YOUR_CLIENT_SECRET"
)

# Use with sync_fetch
doc <- sync_fetch(server$url, "myDocId", token = token, tls = tls)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{validate_token}
\alias{validate_token}
\title{Validate a JWT token}
\usage{
validate_token(
  token,
  issuer,
  client_id,
  allowed_emails = NULL,
  allowed_domains = NULL,
  custom_validator = NULL
)
}
\arguments{
\item{token}{Character string, the JWT.}

\item{issuer}{Expected issuer URL.}

\item{client_id}{Expected audience (client ID).}

\item{allowed_emails}{Character vector of allowed email addresses (optional).}

\item{allowed_domains}{Character vector of allowed email domains (optional).}

\item{custom_validator}{Function(claims) returning TRUE/FALSE (optional).}
}
\value{
List with \code{valid} (logical), \code{email} (character or NULL),
\code{error} (character or NULL).
}
\description{
Verifies a JWT locally using the issuer's public keys (JWKS) and validates
standard OIDC claims.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{oidc_issuer}
\alias{oidc_issuer}
\title{Default OIDC issuer URL}
\usage{
oidc_issuer()
}
\value{
Character string, the OIDC issuer URL.
}
\description{
Returns the \code{OIDC_ISSUER} environment variable if set and non-empty,
otherwise falls back to Google (\code{"https://accounts.google.com"}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autosync-package.R
\docType{package}
\name{autosync-package}
\alias{autosync-package}
\alias{autosync}
\title{autosync: 'Automerge' Sync Server and Client}
\description{
A WebSocket-based implementation of the 'automerge-repo' synchronization
protocol used by 'sync.automerge.org'. Acts as a sync server, enabling 'R' to
serve as a synchronization hub for 'Automerge' clients in 'JavaScript',
'Rust', and other languages, and as a client for fetching, editing, and
synchronizing documents hosted on remote servers.
}
\section{Main Functions}{

\describe{
\item{\code{\link[=sync_server]{sync_server()}}}{Create a new sync server with \verb{$start()} and
\verb{$close()} methods}
}
}

\section{Document Management}{

\describe{
\item{\code{\link[=create_document]{create_document()}}}{Create a new document}
\item{\code{\link[=get_document]{get_document()}}}{Retrieve a document by ID}
\item{\code{\link[=list_documents]{list_documents()}}}{List all document IDs}
\item{\code{\link[=generate_document_id]{generate_document_id()}}}{Generate a new document ID}
}
}

\section{Protocol}{

The server implements the automerge-repo sync protocol over WebSockets.
Messages are CBOR-encoded and include:
\describe{
\item{join/peer}{Handshake messages for connection establishment}
\item{request/sync}{Document synchronization messages}
\item{ephemeral}{Transient messages forwarded without persistence}
\item{error}{Error notifications}
}
}

\section{Example}{

\preformatted{
# Create and start a server
server <- sync_server()
server$start()
server$url

# Stop when done
server$close()
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://posit-dev.github.io/autosync/}
  \item \url{https://github.com/posit-dev/autosync}
  \item Report bugs at \url{https://github.com/posit-dev/autosync/issues}
}

}
\author{
\strong{Maintainer}: Charlie Gao \email{charlie.gao@posit.co} (\href{https://orcid.org/0000-0002-0750-061X}{ORCID})

Authors:
\itemize{
  \item Charlie Gao \email{charlie.gao@posit.co} (\href{https://orcid.org/0000-0002-0750-061X}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/server.R
\name{get_document}
\alias{get_document}
\title{Get a document from the server}
\usage{
get_document(server, doc_id)
}
\arguments{
\item{server}{An autosync_server object.}

\item{doc_id}{Document ID string.}
}
\value{
Automerge document object, or NULL if not found.
}
\description{
Retrieves an Automerge document by its ID.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
server <- sync_server()
doc_id <- create_document(server)
get_document(server, doc_id)
server$close()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{auth_config}
\alias{auth_config}
\title{Create an authentication configuration}
\usage{
auth_config(
  client_id = Sys.getenv("OIDC_CLIENT_ID"),
  issuer = oidc_issuer(),
  allowed_emails = NULL,
  allowed_domains = NULL,
  custom_validator = NULL
)
}
\arguments{
\item{client_id}{The OIDC client ID (application ID). Validated against the
\code{aud} claim in JWTs. Defaults to the \code{OIDC_CLIENT_ID} environment variable.}

\item{issuer}{The OIDC issuer URL. This is used to discover the provider's
public keys via the \code{.well-known/openid-configuration} endpoint, and to
validate the \code{iss} claim in JWTs. Defaults to the \code{OIDC_ISSUER}
environment variable, falling back to Google
(\code{"https://accounts.google.com"}).}

\item{allowed_emails}{Character vector of allowed email addresses. When set,
a token is rejected unless it carries an \code{email} claim with
\code{email_verified} explicitly \code{TRUE}.}

\item{allowed_domains}{Character vector of allowed email domains
(e.g., "mycompany.com"). Subject to the same verified-email requirement as
\code{allowed_emails}.}

\item{custom_validator}{Function(claims) returning TRUE/FALSE for
custom validation logic. Receives the decoded JWT claims as a list.}
}
\value{
An object of class \code{"autosync_auth_config"}.
}
\description{
Creates a configuration object for enabling OIDC JWT authentication
on an autosync server. When enabled, clients must include a valid JWT
(ID token) as a Bearer token in the Authorization header of the WebSocket
upgrade request. Connections without valid credentials are rejected
immediately at connection time.
}
\details{
Works with any OIDC-compliant identity provider: Google, Microsoft Entra,
Okta, Auth0, etc.
}
\examples{
# Google (default issuer)
auth_config(
  client_id = "123456789.apps.googleusercontent.com",
  allowed_domains = "mycompany.com"
)

# Microsoft Entra
auth_config(
  client_id = "abcdef-1234-5678",
  issuer = "https://login.microsoftonline.com/common/v2.0",
  allowed_emails = "alice@mycompany.com"
)

# Custom validator
auth_config(
  client_id = "0oaXXXXXXXX",
  issuer = "https://dev-123456.okta.com/oauth2/default",
  custom_validator = function(claims) "editors" \%in\% claims$groups
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/server.R
\name{print.autosync_server}
\alias{print.autosync_server}
\title{Print method for autosync_server}
\usage{
\method{print}{autosync_server}(x, ...)
}
\arguments{
\item{x}{An autosync_server object.}

\item{...}{Ignored.}
}
\value{
Invisibly returns x.
}
\description{
Print method for autosync_server
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/server.R
\name{sync_server}
\alias{sync_server}
\title{Create an Automerge sync server}
\usage{
sync_server(
  port = 0L,
  host = "127.0.0.1",
  data_dir = tempfile("autosync"),
  auto_create_docs = TRUE,
  storage_id = NULL,
  tls = NULL,
  auth = NULL,
  share = NA
)
}
\arguments{
\item{port}{Port to listen on. Default 0 (binds to a random available port).
The actual URL is retrieved via \code{server$url}.}

\item{host}{Host address to bind to. Default "127.0.0.1" (localhost).}

\item{data_dir}{Directory for document storage. Defaults to a
session-temporary directory, so documents do not persist across R
sessions. Supply an explicit path to persist documents across sessions.}

\item{auto_create_docs}{Logical, whether to auto-create documents when
clients request unknown document IDs. Default TRUE.}

\item{storage_id}{Optional storage ID for this server. If NULL (default),
generates a new persistent identity. Set to NA for an ephemeral server
(no persistence identity).}

\item{tls}{(optional) for secure wss:// connections, a TLS configuration
object created by \code{\link[nanonext:tls_config]{nanonext::tls_config()}}.}

\item{auth}{Optional authentication configuration created by \code{\link[=auth_config]{auth_config()}}.
When provided, clients must include a valid JWT (ID token) as a
Bearer token in the Authorization header of the WebSocket upgrade request.
Connections without valid credentials are rejected immediately.
Note: TLS is required when authentication is enabled to protect tokens.}

\item{share}{Controls document sharing policy for connected clients. This
unified parameter governs both proactive document announcement (pushing
documents to clients) and access control (allowing clients to request
documents). Accepts one of:
\itemize{
\item \code{NA} (default) — never announce but allow all requests.
\item \code{TRUE} — announce all documents to all clients and allow all
requests.
\item \code{FALSE} — never announce and deny all requests (sends
\code{doc-unavailable}).
\item A function with signature \verb{function(client_id, doc_id)}
returning \code{TRUE} (announce and allow), \code{NA} (allow on request only),
or \code{FALSE} (deny access). Called per client and per document.
}}
}
\value{
An autosync_server object inheriting from 'nanoServer', with
\verb{$start()} and \verb{$close()} methods.
}
\description{
Creates a WebSocket server that implements the automerge-repo sync protocol,
compatible with JavaScript, Rust, and other Automerge clients.
}
\details{
The returned server inherits from nanonext's nanoServer class and provides
\verb{$start()} and \verb{$close()} methods for non-blocking operation.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
# Create and start a server
server <- sync_server()
server$start()

# Server is now running in the background
# ...do other work...

# Stop when done
server$close()

# With TLS for secure connections
cert <- nanonext::write_cert()
tls <- nanonext::tls_config(server = cert$server)
server <- sync_server(tls = tls)
server$start()
server$url
server$close()

# Server with OIDC authentication (requires TLS)
cert <- nanonext::write_cert()
tls <- nanonext::tls_config(server = cert$server)
server <- sync_server(
  tls = tls,
  auth = auth_config(
    client_id = "123456789.apps.googleusercontent.com",
    allowed_domains = "mycompany.com"
  )
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/server.R
\name{create_document}
\alias{create_document}
\title{Create a new document on the server}
\usage{
create_document(server, doc_id = NULL)
}
\arguments{
\item{server}{An autosync_server object.}

\item{doc_id}{Optional document ID. If NULL, generates a new ID.}
}
\value{
Document ID string.
}
\description{
Creates a new empty Automerge document and registers it with the server.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
server <- sync_server()
doc_id <- create_document(server)
server$close()
\dontshow{\}) # examplesIf}
}
